import argparse
import asyncio
import base64
import configparser
import hashlib
import json
import os
import time
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse
import pathlib
import websockets
from nacl.signing import SigningKey
from config import APP_DIR


NODE_SIGN_KEY_PATH = APP_DIR / "node_sign_ed25519.key"

def b64e(b: bytes) -> str:
    return base64.b64encode(b).decode("ascii")

def b64d(s: str) -> bytes:
    return base64.b64decode(s.encode("ascii"))

def load_or_create_node_signing_key(path: Path = NODE_SIGN_KEY_PATH):
    if SigningKey is None:
        raise RuntimeError(
            "PyNaCl is required for node AUTH but isn't available in this Python environment. "
            "Install it for /usr/bin/python3 (e.g. `python3 -m pip install --user PyNaCl` or distro package `python3-nacl`) "
            "or update your systemd service to use the venv that has PyNaCl."
        )

    try:
        if path.exists():
            seed_b64 = path.read_text(encoding="utf-8").strip()
            seed = b64d(seed_b64)
            if len(seed) != 32:
                raise ValueError("bad seed length")
            return SigningKey(seed)
    except Exception:
        try:
            if path.exists():
                bad = path.with_suffix(path.suffix + f".bad.{now_ts()}")
                path.rename(bad)
        except Exception:
            pass

    sk = SigningKey.generate()
    try:
        seed = sk.encode()
        path.write_text(b64e(seed) + "\n", encoding="utf-8")
        try:
            os.chmod(path, 0o600)
        except Exception:
            pass
    except Exception:
        pass
    return sk

def canonical_json(obj) -> str:
    return json.dumps(obj, separators=(",", ":"), sort_keys=True, ensure_ascii=False)

def now_ts() -> int:
    return int(time.time())

def compute_file_id_from_container(blob: bytes) -> Optional[str]:

    try:
        h = hashlib.sha256()
        off = 0
        n = len(blob)
        while off + 4 <= n:
            ln = int.from_bytes(blob[off : off + 4], "big")
            off += 4
            if ln <= 0 or ln > 64 * 1024 * 1024:
                return None
            if off + ln > n:
                return None
            rec = blob[off : off + ln]
            off += ln
            h.update(rec)

        if off != n:
            return None
        return h.hexdigest()
    except Exception:
        return None

def _parse_bool(v, default: bool = False) -> bool:
    if isinstance(v, bool):
        return v
    if v is None:
        return default
    try:
        s = str(v).strip().lower()
    except Exception:
        return default
    if s in ("1", "true", "yes", "y", "on"):
        return True
    if s in ("0", "false", "no", "n", "off"):
        return False
    return default

def load_conf(path: Path, default: dict) -> dict:
    try:
        text = path.read_text(encoding="utf-8")
    except Exception:
        return default

    def _strip_wrapping_quotes(v: str) -> str:
        if not isinstance(v, str):
            return str(v)
        s = v.strip()
        if len(s) >= 2 and ((s.startswith('"') and s.endswith('"')) or (s.startswith("'") and s.endswith("'"))):
            return s[1:-1]
        return s

    try:
        parser = configparser.ConfigParser(
            interpolation=None,
            delimiters=("=",),
            comment_prefixes=("#", ";"),
            inline_comment_prefixes=("#", ";"),
            strict=False,
            default_section="DEFAULT",
        )

        first_content_line = None
        for raw in text.splitlines():
            s = raw.strip()
            if not s:
                continue
            if s.startswith(("#", ";")):
                continue
            first_content_line = s
            break

        if first_content_line and not first_content_line.startswith("["):
            parse_text = "[DEFAULT]\n" + text
        else:
            parse_text = text

        parser.read_string(parse_text)

        out: dict = {}

        try:
            for k, v in parser.defaults().items():
                ks = str(k).strip()
                if not ks:
                    continue
                out[ks] = _strip_wrapping_quotes(str(v))
        except Exception:
            pass

        for sec in parser.sections():
            sec_name = str(sec).strip().lower()
            sec_items: dict[str, str] = {}
            try:
                for k, v in parser.items(sec, raw=True):
                    ks = str(k).strip()
                    if not ks:
                        continue

                    if ks in parser.defaults():
                        continue
                    sec_items[ks] = _strip_wrapping_quotes(str(v))
            except Exception:
                continue

            if sec_name in ("network", "net"):
                out["network"] = sec_items
            else:
                out[sec_name] = sec_items

        return out if isinstance(out, dict) else default
    except Exception:
        return default

def load_any_config(path: Path, default: dict) -> dict:
    if not path or not isinstance(path, Path):
        return default
    if not path.exists():
        return default
    return load_conf(path, default)

NODE_CONF_DOCS = """# Antch Node Configuration (node.conf)
#
# Format:
# - key = value
# - Comments: '#' or ';'
# - Sections: [network]
#
# This file configures the node (seed). Use [network] for connection settings.
#
# ----------------------------
# Keys
# ----------------------------
# node_id
#   Node identity string used when connecting to server.
#   Example: node_id = MY-NODE-1
#
# data_dir
#   Directory where blobs/metas are stored.
#   Example: data_dir = ~/.antch/node_store
#
# max_storage_mb
#   Global storage budget (MB). Policy controlled by on_storage_exceed.
#   Example: max_storage_mb = 102400
#
# max_blob_mb
#   Per-blob safety limit (MB) for any kind.
#
# min_file_blob_mb / max_file_blob_mb
#   For kind=file only: minimum/maximum size per file blob (MB).
#
# max_user_file_storage_mb
#   Per-owner total quota for file blobs (MB). 0 disables.
#
# on_user_quota_exceed
#   When per-owner quota would be exceeded: reject | evict_oldest
#
# on_storage_exceed
#   When global storage would be exceeded: reject | evict_oldest
#
# connect_timeout_sec
#   WebSocket connect timeout (seconds).
#
# seed_chat_owners
#   Restrict db snapshot seeding to a comma-separated list. Empty means ALL.
#   Example: seed_chat_owners = alice,bob
#
# seed_files
#   Enable/disable storing kind=file blobs on disk. If false, the node won't store files.
#   Example: seed_files = false
#
# [network] section
# server_url
#   Server WebSocket URL (ws:// or wss://).
# mode
#   direct | proxy (SOCKS5)
# proxy_host / proxy_port / proxy_rdns
#   SOCKS5 settings. proxy_rdns is true/false.
"""

def default_node_conf_text() -> str:
    cfg = default_node_config()
    net = cfg.get("network") if isinstance(cfg.get("network"), dict) else {}

    def _b(v: bool) -> str:
        return "true" if bool(v) else "false"

    return (
        NODE_CONF_DOCS
        + "\n"
        + f"node_id = {cfg.get('node_id')}\n"
        + f"node_description = {cfg.get('node_description', '')}\n"
        + f"data_dir = {cfg.get('data_dir')}\n"
        + f"max_storage_mb = {cfg.get('max_storage_mb')}\n"
        + f"max_blob_mb = {cfg.get('max_blob_mb')}\n"
        + "\n"
        + "# File seeding limits (kind=file)\n"
        + f"seed_files = {str(cfg.get('seed_files', False)).lower()}\n"
        + f"min_file_blob_mb = {cfg.get('min_file_blob_mb')}\n"
        + f"max_file_blob_mb = {cfg.get('max_file_blob_mb')}\n"
        + f"max_user_file_storage_mb = {cfg.get('max_user_file_storage_mb')}\n"
        + f"on_user_quota_exceed = {cfg.get('on_user_quota_exceed')}\n"
        + f"on_storage_exceed = {cfg.get('on_storage_exceed')}\n"
        + "\n"
        + f"connect_timeout_sec = {cfg.get('connect_timeout_sec')}\n"
        + "# seed_chat_owners = alice,bob   # empty means ALL\n"
        + "\n"
        + "[network]\n"
        + f"server_url = {net.get('server_url')}\n"
        + f"mode = {net.get('mode')}\n"
        + f"proxy_host = {net.get('proxy_host')}\n"
        + f"proxy_port = {net.get('proxy_port')}\n"
        + f"proxy_rdns = {_b(net.get('proxy_rdns', True))}\n"
    )

def default_node_config() -> dict:
    return {
        "node_id": f"node-{os.uname().nodename}",
        "node_description": "",
        "network": {
            "server_url": "wss://antch.lama2923.com",

            "mode": "direct",
            "proxy_host": "127.0.0.1",
            "proxy_port": 9050,
            "proxy_rdns": True,
        },
        "data_dir": str(APP_DIR / "node_store"),
        "seed_files": True,
        "max_storage_mb": 102400,

        "max_blob_mb": 64,

        "min_file_blob_mb": 0,
        "max_file_blob_mb": 64,

        "max_user_file_storage_mb": 0,

        "on_user_quota_exceed": "reject",

        "on_storage_exceed": "evict_oldest",
        "connect_timeout_sec": 10,
    }

def ensure_node_config(node_config_path: Path) -> Path:
    try:
        node_config_path.parent.mkdir(parents=True, exist_ok=True)
    except Exception:
        pass

    if node_config_path.exists():
        return node_config_path

    fallback_conf = APP_DIR / "node.conf"
    if fallback_conf.exists():
        return fallback_conf

    conf_text = default_node_conf_text()
    try:
        node_config_path.write_text(conf_text + ("\n" if not conf_text.endswith("\n") else ""), encoding="utf-8")
    except Exception:
        return node_config_path

    example_path = APP_DIR / "node_config.example.conf"
    try:
        if not example_path.exists():
            example_path.write_text(conf_text + ("\n" if not conf_text.endswith("\n") else ""), encoding="utf-8")
    except Exception:
        pass

    print(f"Created node config: {node_config_path}")
    print(f"Example config: {example_path}")
    return node_config_path

@dataclass
class NodeConfig:
    server_url: str
    node_id: str
    node_description: str
    data_dir: Path
    max_storage_mb: int
    max_blob_mb: int
    min_file_blob_mb: int
    max_file_blob_mb: int
    max_user_file_storage_mb: int
    on_user_quota_exceed: str
    on_storage_exceed: str
    connect_timeout_sec: int
    proxy_mode: str
    proxy_host: str
    proxy_port: int
    proxy_rdns: bool
    seed_chat_owners: list[str]
    seed_files: bool

    @staticmethod
    def from_file(node_config_path: Path) -> "NodeConfig":
        node_cfg = load_any_config(node_config_path, {})
        node_net = node_cfg.get("network") if isinstance(node_cfg.get("network"), dict) else {}

        server_url = str(node_net.get("server_url") or node_cfg.get("server_url") or "ws://127.0.0.1:8765")

        node_id = str(node_cfg.get("node_id") or f"node-{os.uname().nodename}")

        node_description = ""
        try:
            node_description = str(node_cfg.get("node_description") or "").strip()
        except Exception:
            node_description = ""
        if len(node_description) > 200:
            node_description = node_description[:200]

        raw_data_dir = node_cfg.get("data_dir") or str(APP_DIR / "node_store")
        if isinstance(raw_data_dir, str):
            raw_data_dir = os.path.expanduser(raw_data_dir)
        data_dir = Path(raw_data_dir)
        max_storage_mb = int(node_cfg.get("max_storage_mb") or 102400)
        max_blob_mb = int(node_cfg.get("max_blob_mb") or 64)

        def _int_or(default: int, v) -> int:
            try:
                return int(v)
            except Exception:
                return int(default)

        min_file_blob_mb = _int_or(0, node_cfg.get("min_file_blob_mb"))
        max_file_blob_mb = _int_or(max_blob_mb, node_cfg.get("max_file_blob_mb"))
        if max_file_blob_mb <= 0:
            max_file_blob_mb = max_blob_mb
        if min_file_blob_mb < 0:
            min_file_blob_mb = 0
        if max_file_blob_mb < min_file_blob_mb:
            max_file_blob_mb = min_file_blob_mb

        max_user_file_storage_mb = _int_or(0, node_cfg.get("max_user_file_storage_mb"))
        if max_user_file_storage_mb < 0:
            max_user_file_storage_mb = 0

        def _policy(v, default: str) -> str:
            try:
                s = str(v or "").strip().lower()
            except Exception:
                s = ""
            if s in ("reject", "evict_oldest"):
                return s
            return default

        on_user_quota_exceed = _policy(node_cfg.get("on_user_quota_exceed"), "reject")
        on_storage_exceed = _policy(node_cfg.get("on_storage_exceed"), "evict_oldest")

        connect_timeout_sec = int(node_cfg.get("connect_timeout_sec") or 10)

        proxy_mode = str(node_net.get("mode") or node_cfg.get("proxy_mode") or "direct").strip().lower()
        proxy_host = str(node_net.get("proxy_host") or node_cfg.get("proxy_host") or "127.0.0.1").strip()
        try:
            proxy_port = int(node_net.get("proxy_port") or node_cfg.get("proxy_port") or 9050)
        except Exception:
            proxy_port = 9050
        proxy_rdns = _parse_bool(node_net.get("proxy_rdns"), True)

        seed_chat_owners: list[str] = []
        try:
            raw = node_cfg.get("seed_chat_owners")
            if isinstance(raw, list):
                seed_chat_owners = [str(x).strip() for x in raw if isinstance(x, str) and str(x).strip()]
            elif isinstance(raw, str):
                seed_chat_owners = [s.strip() for s in raw.split(",") if s.strip()]
        except Exception:
            seed_chat_owners = []

        seed_files = True
        try:
            v = node_cfg.get("seed_files")
            if isinstance(v, bool):
                seed_files = bool(v)
            elif isinstance(v, (int, float)):
                seed_files = bool(int(v))
            elif isinstance(v, str):
                seed_files = v.strip().lower() in ("1", "true", "yes", "y", "on")
        except Exception:
            seed_files = True

        return NodeConfig(
            server_url=server_url,
            node_id=node_id,
            node_description=node_description,
            data_dir=data_dir,
            max_storage_mb=max_storage_mb,
            max_blob_mb=max_blob_mb,
            min_file_blob_mb=min_file_blob_mb,
            max_file_blob_mb=max_file_blob_mb,
            max_user_file_storage_mb=max_user_file_storage_mb,
            on_user_quota_exceed=on_user_quota_exceed,
            on_storage_exceed=on_storage_exceed,
            connect_timeout_sec=connect_timeout_sec,
            proxy_mode=proxy_mode,
            proxy_host=proxy_host,
            proxy_port=proxy_port,
            proxy_rdns=proxy_rdns,
            seed_chat_owners=seed_chat_owners,
            seed_files=seed_files,
        )

class BlobStore:
    def __init__(self, root: Path, max_storage_mb: int):
        self.root = root
        self.blobs_dir = root / "blobs"
        self.meta_dir = root / "meta"
        self.max_bytes = int(max_storage_mb) * 1024 * 1024
        self.blobs_dir.mkdir(parents=True, exist_ok=True)
        self.meta_dir.mkdir(parents=True, exist_ok=True)

    def _blob_path(self, sha256_hex: str) -> Path:
        return self.blobs_dir / f"{sha256_hex}.bin"

    def _meta_path(self, sha256_hex: str) -> Path:
        return self.meta_dir / f"{sha256_hex}.json"

    def _owner_key(self, owner: str) -> str:
        return hashlib.sha256(owner.encode("utf-8")).hexdigest()

    def _latest_db_path(self, owner: str) -> Path:
        return self.meta_dir / f"latest_db_{self._owner_key(owner)}.json"

    def get_latest_db(self, owner: str) -> Optional[dict]:
        p = self._latest_db_path(owner)
        if not p.exists():
            return None
        try:
            v = json.loads(p.read_text(encoding="utf-8"))
        except Exception:
            return None
        if not isinstance(v, dict):
            return None
        return v

    def set_latest_db(self, owner: str, blob_sha256: str, created_at: int) -> None:
        p = self._latest_db_path(owner)
        data = {
            "owner": str(owner),
            "kind": "db",
            "blob_sha256": str(blob_sha256),
            "created_at": int(created_at),
            "updated_at": now_ts(),
        }
        try:
            p.write_text(canonical_json(data), encoding="utf-8")
        except Exception:
            pass

    def delete(self, sha256_hex: str) -> None:
        try:
            self._blob_path(sha256_hex).unlink(missing_ok=True)
        except Exception:
            pass
        try:
            self._meta_path(sha256_hex).unlink(missing_ok=True)
        except Exception:
            pass

    def clear_latest_db_if_matches(self, owner: str, blob_sha256: str) -> None:
        try:
            st = self.get_latest_db(owner)
            if not isinstance(st, dict):
                return
            if st.get("blob_sha256") != blob_sha256:
                return
            p = self._latest_db_path(owner)
            p.unlink(missing_ok=True)
        except Exception:
            pass

    def has(self, sha256_hex: str) -> bool:
        return self._blob_path(sha256_hex).exists()

    def put(self, sha256_hex: str, blob: bytes, meta: dict):
        p = self._blob_path(sha256_hex)
        if p.exists():
            try:
                self._meta_path(sha256_hex).write_text(canonical_json(meta), encoding="utf-8")
            except Exception:
                pass
            return
        tmp = p.with_suffix(".tmp")
        tmp.write_bytes(blob)
        os.replace(tmp, p)
        try:
            self._meta_path(sha256_hex).write_text(canonical_json(meta), encoding="utf-8")
        except Exception:
            pass

    def get(self, sha256_hex: str) -> Optional[bytes]:
        p = self._blob_path(sha256_hex)
        if not p.exists():
            return None
        try:
            return p.read_bytes()
        except Exception:
            return None

    def get_meta(self, sha256_hex: str) -> Optional[dict]:
        p = self._meta_path(sha256_hex)
        if not p.exists():
            return None
        try:
            v = json.loads(p.read_text(encoding="utf-8"))
        except Exception:
            return None
        if not isinstance(v, dict):
            return None
        return v

    def ensure_file_id_for_sha(self, sha256_hex: str) -> Optional[str]:
        meta = self.get_meta(sha256_hex) or {}
        if not isinstance(meta, dict):
            meta = {}
        fid = meta.get("file_id")
        if isinstance(fid, str) and len(fid) == 64:
            return fid
        blob = self.get(sha256_hex)
        if blob is None:
            return None
        derived = compute_file_id_from_container(blob)
        if not isinstance(derived, str) or len(derived) != 64:
            return None

        meta["file_id"] = derived
        meta["file_id_source"] = "derived"
        try:
            self._meta_path(sha256_hex).write_text(canonical_json(meta), encoding="utf-8")
        except Exception:
            pass
        return derived

    def total_bytes(self) -> int:
        total = 0
        try:
            for f in self.blobs_dir.glob("*.bin"):
                try:
                    total += f.stat().st_size
                except Exception:
                    pass
        except Exception:
            return 0
        return total

    def cleanup_if_needed(self):
        total = self.total_bytes()
        if total <= self.max_bytes:
            return

        files = []
        for f in self.blobs_dir.glob("*.bin"):
            try:
                files.append((f.stat().st_mtime, f))
            except Exception:
                pass
        files.sort(key=lambda x: x[0])

        for _, f in files:
            if total <= self.max_bytes:
                break
            try:
                sz = f.stat().st_size
            except Exception:
                sz = 0
            try:
                f.unlink(missing_ok=True)
            except Exception:
                pass
            meta = self.meta_dir / (f.stem + ".json")
            try:
                meta.unlink(missing_ok=True)
            except Exception:
                pass
            total -= sz

    def _iter_meta_rows(self):
        try:
            for mp in self.meta_dir.glob("*.json"):
                stem = mp.stem
                if len(stem) != 64:
                    continue
                try:
                    meta = json.loads(mp.read_text(encoding="utf-8"))
                except Exception:
                    continue
                if not isinstance(meta, dict):
                    continue
                yield stem, meta
        except Exception:
            return

    def find_file_sha_by_file_id(self, file_id: str) -> Optional[str]:
        if not isinstance(file_id, str) or not file_id or len(file_id) != 64:
            return None
        best_sha = None
        best_ts = -1
        for sha, meta in self._iter_meta_rows():
            try:
                if meta.get("kind") != "file":
                    continue
                fid = meta.get("file_id")
                if fid != file_id:
                    if not fid:
                        fid2 = self.ensure_file_id_for_sha(sha)
                        if fid2 != file_id:
                            continue
                    else:
                        continue
                ts = int(meta.get("stored_at") or 0)
            except Exception:
                continue
            if ts > best_ts:
                best_ts = ts
                best_sha = sha
        return best_sha

    def total_bytes_for_owner_kind(self, owner: str, kind: str) -> int:
        if not isinstance(owner, str) or not owner or kind not in ("db", "file"):
            return 0
        total = 0
        for _, meta in self._iter_meta_rows():
            try:
                if meta.get("owner") != owner or meta.get("kind") != kind:
                    continue
                total += int(meta.get("bytes_total") or 0)
            except Exception:
                continue
        return int(total)

    def evict_oldest_for_owner_kind(self, owner: str, kind: str, bytes_to_free: int) -> int:
        if not isinstance(owner, str) or not owner or kind not in ("db", "file"):
            return 0
        try:
            need = int(bytes_to_free)
        except Exception:
            need = 0
        if need <= 0:
            return 0

        cands: list[tuple[int, str, int]] = []
        for sha, meta in self._iter_meta_rows():
            try:
                if meta.get("owner") != owner or meta.get("kind") != kind:
                    continue
                stored_at = int(meta.get("stored_at") or 0)
                sz = int(meta.get("bytes_total") or 0)
                cands.append((stored_at, sha, sz))
            except Exception:
                continue
        cands.sort(key=lambda x: x[0])

        freed = 0
        for _, sha, sz in cands:
            if freed >= need:
                break
            self.delete(sha)
            try:
                freed += int(sz)
            except Exception:
                pass
        return int(freed)

class NodeClient:
    def __init__(self, cfg: NodeConfig, store: BlobStore):
        self.cfg = cfg
        self.store = store
        self._max_blob_bytes = max(1, int(cfg.max_blob_mb)) * 1024 * 1024
        self._min_file_blob_bytes = max(0, int(cfg.min_file_blob_mb)) * 1024 * 1024
        self._max_file_blob_bytes = max(1, int(cfg.max_file_blob_mb)) * 1024 * 1024
        self._debug = str(os.getenv("ANTCH_NODE_DEBUG") or "").strip().lower() in ("1", "true", "yes", "y", "on")
        self._sign_sk = load_or_create_node_signing_key()
        self._sign_pub_b64 = b64e(bytes(self._sign_sk.verify_key))
        self._chunked_uploads: dict[str, dict] = {}

    def _log(self, msg: str) -> None:
        if not self._debug:
            return
        try:
            print(f"[node][debug] {msg}", flush=True)
        except Exception:
            pass

    async def run_forever(self):
        backoff = 1.0
        while True:
            try:
                await self._run_once()
                backoff = 1.0
            except Exception as e:
                try:
                    print(f"[node] connection error: {e}", flush=True)
                except Exception:
                    pass
                await asyncio.sleep(backoff)
                backoff = min(backoff * 2.0, 30.0)

    async def _run_once(self):
        try:
            print(f"[node] starting: node_id={self.cfg.node_id} server={self.cfg.server_url} mode={self.cfg.proxy_mode}", flush=True)
            print(f"[node] store_dir={self.cfg.data_dir} max_storage_mb={self.cfg.max_storage_mb} max_blob_mb={self.cfg.max_blob_mb}", flush=True)
            print(
                f"[node] file_limits=min={self.cfg.min_file_blob_mb}MB max={self.cfg.max_file_blob_mb}MB user_quota={self.cfg.max_user_file_storage_mb}MB on_user_quota_exceed={self.cfg.on_user_quota_exceed} on_storage_exceed={self.cfg.on_storage_exceed}",
                flush=True,
            )
            print(f"[node] seed_files={bool(self.cfg.seed_files)}", flush=True)
            if self.cfg.seed_chat_owners:
                print(f"[node] seed_chat_owners={self.cfg.seed_chat_owners}", flush=True)
            else:
                print("[node] seed_chat_owners=ALL", flush=True)
        except Exception:
            pass

        connect_args = {
            "ping_interval": 20,
            "ping_timeout": 120,
            "open_timeout": self.cfg.connect_timeout_sec,
        }

        try:
            connect_args["max_size"] = 256 * 1024 * 1024
        except Exception:
            pass

        try:
            if not str(self.cfg.server_url).startswith(("ws://", "wss://")):
                raise RuntimeError(f"server_url must be ws:// or wss:// (got {self.cfg.server_url!r})")
        except Exception as e:
            raise

        if self.cfg.proxy_mode in ("proxy", "tor", "socks5"):
            try:
                import socks
            except Exception as e:
                raise RuntimeError("PySocks not installed but proxy mode is enabled") from e

            u = urlparse(self.cfg.server_url)
            host = u.hostname
            port = u.port or (443 if u.scheme == "wss" else 80)
            if host:
                loop = asyncio.get_running_loop()

                proxy_port = int(self.cfg.proxy_port)
                if proxy_port <= 0 or proxy_port > 65535:
                    raise RuntimeError(f"Invalid proxy_port={proxy_port}; expected 1..65535")

                def create_proxy_socket():
                    s = socks.socksocket()
                    s.set_proxy(socks.SOCKS5, self.cfg.proxy_host, int(proxy_port), rdns=bool(self.cfg.proxy_rdns))
                    s.settimeout(int(self.cfg.connect_timeout_sec))
                    s.connect((host, int(port)))
                    return s

                connect_args["sock"] = await loop.run_in_executor(None, create_proxy_socket)

        async with websockets.connect(self.cfg.server_url, **connect_args) as ws:
            await ws.send(
                canonical_json(
                    {
                        "type": "node_hello",
                        "node_id": self.cfg.node_id,
                        "description": self.cfg.node_description,
                        "sign_pub": self._sign_pub_b64,
                        "caps": {
                            "store": True,
                            "fetch": True,
                            "max_storage_mb": int(self.cfg.max_storage_mb),
                            "max_blob_mb": int(self.cfg.max_blob_mb),
                        },
                    }
                )
            )

            try:
                raw = await ws.recv()
            except Exception:
                return
            try:
                msg = json.loads(raw)
            except Exception:
                return

            if msg.get("type") == "node_challenge":
                nonce = msg.get("nonce")
                srv_ts = msg.get("ts")
                if not isinstance(nonce, str) or not nonce:
                    return
                try:
                    srv_ts_i = int(srv_ts)
                except Exception:
                    return
                payload = {"node_id": str(self.cfg.node_id), "nonce": str(nonce), "ts": int(srv_ts_i)}
                to_sign = ("node_auth|" + canonical_json(payload)).encode("utf-8")
                try:
                    sig = self._sign_sk.sign(to_sign).signature
                except Exception:
                    return
                await ws.send(
                    canonical_json(
                        {
                            "type": "node_auth",
                            "node_id": str(self.cfg.node_id),
                            "nonce": str(nonce),
                            "ts": int(srv_ts_i),
                            "sign_pub": self._sign_pub_b64,
                            "sig": b64e(sig),
                        }
                    )
                )

                try:
                    raw2 = await ws.recv()
                    msg2 = json.loads(raw2)
                except Exception:
                    return
                if msg2.get("type") != "node_ok":
                    return
            elif msg.get("type") != "node_ok":
                return

            try:
                print("[node] connected (node_ok)", flush=True)
            except Exception:
                pass

            async def _alive_loop():
                while True:
                    try:
                        await asyncio.sleep(15)
                    except Exception:
                        return
                    try:
                        await ws.send(canonical_json({"type": "node_alive", "ts": now_ts()}))
                    except Exception:
                        return

            asyncio.create_task(_alive_loop())

            while True:
                raw = await ws.recv()
                try:
                    msg = json.loads(raw)
                except Exception:
                    continue

                t = msg.get("type")
                if t == "node_store":
                    await self._handle_store(ws, msg)
                elif t == "node_store_begin":
                    await self._handle_store_begin(ws, msg)
                elif t == "node_store_chunk":
                    await self._handle_store_chunk(ws, msg)
                elif t == "node_store_end":
                    await self._handle_store_end(ws, msg)
                elif t == "node_fetch":
                    await self._handle_fetch(ws, msg)
                elif t == "node_fetch_file":
                    await self._handle_fetch_file(ws, msg)
                elif t == "node_delete":
                    await self._handle_delete(msg)
                elif t == "node_db_candidates_req":
                    await self._handle_db_candidates(ws, msg)
                elif t == "node_db_has_newer_req":
                    await self._handle_db_has_newer(ws, msg)

    async def _handle_delete(self, msg: dict) -> None:
        blob_sha = msg.get("blob_sha256")
        kind = msg.get("kind")
        owner = msg.get("owner")
        if not isinstance(blob_sha, str) or not blob_sha:
            return
        if kind not in ("db", "file"):
            return

        if kind == "db" and isinstance(owner, str) and owner:
            self.store.clear_latest_db_if_matches(owner, blob_sha)

        self.store.delete(blob_sha)
        try:
            print(f"[node] deleted blob {blob_sha[:12]}.. kind={kind} owner={owner}", flush=True)
        except Exception:
            pass

    async def _handle_store(self, ws, msg: dict):
        blob_sha = msg.get("blob_sha256")
        ct = msg.get("ct")
        kind = msg.get("kind")
        owner = msg.get("owner")
        created_at = msg.get("created_at")
        sig = msg.get("sig")
        file_id = msg.get("file_id")
        if not isinstance(blob_sha, str) or not isinstance(ct, str):
            return

        if kind not in ("db", "file"):
            return

        if kind == "file" and not bool(getattr(self.cfg, "seed_files", True)):
            return

        if kind == "file":
            if str(self.cfg.on_storage_exceed).strip().lower() == "reject":
                try:
                    if (self.store.total_bytes() + int(msg.get("bytes_total") or 0)) > int(self.store.max_bytes):
                        return
                except Exception:
                    pass

        if kind == "db":
            if not isinstance(owner, str) or not owner.strip():
                return
            if self.cfg.seed_chat_owners and owner not in self.cfg.seed_chat_owners:
                return

        if len(ct) > int(self._max_blob_bytes * 1.5) + 16:
            return

        try:
            blob = b64d(ct)
        except Exception:
            return

        if len(blob) > self._max_blob_bytes:
            return

        if kind == "file":
            if self._min_file_blob_bytes and len(blob) < self._min_file_blob_bytes:
                return
            if self._max_file_blob_bytes and len(blob) > self._max_file_blob_bytes:
                return

            if str(self.cfg.on_storage_exceed).strip().lower() == "reject":
                if (self.store.total_bytes() + len(blob)) > int(self.store.max_bytes):
                    return

            if isinstance(owner, str) and owner and int(self.cfg.max_user_file_storage_mb or 0) > 0:
                limit_bytes = int(self.cfg.max_user_file_storage_mb) * 1024 * 1024
                cur = self.store.total_bytes_for_owner_kind(owner, "file")
                if (cur + len(blob)) > limit_bytes:
                    pol = str(self.cfg.on_user_quota_exceed).strip().lower()
                    if pol == "evict_oldest":
                        need = (cur + len(blob)) - limit_bytes
                        self.store.evict_oldest_for_owner_kind(owner, "file", need)
                        cur2 = self.store.total_bytes_for_owner_kind(owner, "file")
                        if (cur2 + len(blob)) > limit_bytes:
                            return
                    else:
                        return

        sha = hashlib.sha256(blob).hexdigest()
        if sha != blob_sha:
            return

        derived_file_id = None
        file_id_source = None
        if kind == "file":
            derived_file_id = compute_file_id_from_container(blob)
            if not isinstance(derived_file_id, str) or len(derived_file_id) != 64:
                return
            if isinstance(file_id, str) and file_id:
                if str(file_id) != derived_file_id:
                    return
                file_id_source = "provided"
            else:
                file_id = derived_file_id
                file_id_source = "derived"

        try:
            created_at_i = int(created_at) if isinstance(created_at, int) else int(created_at or now_ts())
        except Exception:
            created_at_i = now_ts()

        old_sha = None
        old_created = 0
        if kind == "db" and isinstance(owner, str) and owner:
            st = self.store.get_latest_db(owner)
            if isinstance(st, dict):
                try:
                    old_sha = st.get("blob_sha256")
                    old_created = int(st.get("created_at") or 0)
                except Exception:
                    old_sha = None
                    old_created = 0

            if old_sha and old_created and created_at_i <= old_created:
                return

        meta = {
            "owner": owner,
            "kind": kind,
            "created_at": created_at_i,
            "blob_sha256": blob_sha,
            "bytes_total": len(blob),
            "stored_at": now_ts(),
        }

        if kind == "file" and isinstance(file_id, str) and file_id:
            meta["file_id"] = str(file_id)
            if file_id_source:
                meta["file_id_source"] = str(file_id_source)

        if isinstance(sig, str) and sig:
            meta["sig_b64"] = sig

        self.store.put(blob_sha, blob, meta)

        if kind == "db" and isinstance(owner, str) and owner:
            self.store.set_latest_db(owner, blob_sha, created_at_i)
            if old_sha and isinstance(old_sha, str) and old_sha != blob_sha:
                self.store.delete(old_sha)

        if str(self.cfg.on_storage_exceed).strip().lower() == "evict_oldest":
            self.store.cleanup_if_needed()

        try:
            print(
                f"[node] stored blob {blob_sha[:12]}.. bytes={len(blob)} kind={meta.get('kind')} owner={meta.get('owner')}",
                flush=True,
            )
        except Exception:
            pass

    async def _handle_store_begin(self, ws, msg: dict) -> None:
        upload_id = msg.get("upload_id")
        if not isinstance(upload_id, str) or not upload_id:
            return
        self._chunked_uploads[upload_id] = {
            "owner": msg.get("owner"),
            "kind": msg.get("kind"),
            "created_at": msg.get("created_at"),
            "blob_sha256": msg.get("blob_sha256"),
            "file_id": msg.get("file_id"),
            "bytes_total": msg.get("bytes_total"),
            "nonce": msg.get("nonce"),
            "sig": msg.get("sig"),
            "owner_sign_pub": msg.get("owner_sign_pub"),
            "chunks": {},
        }

    async def _handle_store_chunk(self, ws, msg: dict) -> None:
        upload_id = msg.get("upload_id")
        if not isinstance(upload_id, str) or upload_id not in self._chunked_uploads:
            return
        st = self._chunked_uploads[upload_id]
        idx = msg.get("idx")
        data = msg.get("data")
        if not isinstance(idx, int) or not isinstance(data, str):
            return
        try:
            st["chunks"][idx] = b64d(data)
        except Exception:
            pass

    async def _handle_store_end(self, ws, msg: dict) -> None:
        upload_id = msg.get("upload_id")
        if not isinstance(upload_id, str) or upload_id not in self._chunked_uploads:
            return
        st = self._chunked_uploads.pop(upload_id)
        chunks = st.get("chunks", {})
        if not chunks:
            return
        buf = bytearray()
        for idx in sorted(chunks.keys()):
            buf.extend(chunks[idx])
        fake_msg = {
            "blob_sha256": st.get("blob_sha256"),
            "ct": b64e(bytes(buf)),
            "kind": st.get("kind"),
            "owner": st.get("owner"),
            "created_at": st.get("created_at"),
            "sig": st.get("sig"),
            "file_id": st.get("file_id"),
            "bytes_total": st.get("bytes_total"),
        }
        await self._handle_store(ws, fake_msg)

    def _list_db_items(self, owner: str, limit: int = 10, since_created_at: int | None = None) -> list[dict]:
        if not isinstance(owner, str) or not owner:
            return []
        try:
            lim = int(limit)
        except Exception:
            lim = 10
        if lim <= 0:
            lim = 10
        lim = min(lim, 50)

        since = None
        if since_created_at is not None:
            try:
                since = int(since_created_at)
            except Exception:
                since = None

        items: list[dict] = []
        for sha, meta in self.store._iter_meta_rows():
            if not isinstance(meta, dict):
                continue
            if meta.get("kind") != "db" or meta.get("owner") != owner:
                continue
            try:
                created_at = int(meta.get("created_at") or 0)
                bytes_total = int(meta.get("bytes_total") or 0)
            except Exception:
                continue
            if created_at <= 0 or bytes_total <= 0:
                continue
            if since is not None and created_at <= since:
                continue
            sig_b64 = meta.get("sig_b64")
            if not isinstance(sig_b64, str) or not sig_b64:
                continue
            items.append(
                {
                    "blob_sha256": sha,
                    "created_at": int(created_at),
                    "bytes_total": int(bytes_total),
                    "sig": sig_b64,
                }
            )

        items.sort(key=lambda x: int(x.get("created_at") or 0), reverse=True)
        return items[:lim]

    async def _handle_db_candidates(self, ws, msg: dict) -> None:
        req_id = msg.get("req_id")
        owner = msg.get("owner")
        limit = msg.get("limit")
        if not isinstance(req_id, str) or not isinstance(owner, str):
            return
        try:
            lim = int(limit)
        except Exception:
            lim = 10
        items = self._list_db_items(owner, limit=lim)
        try:
            await ws.send(canonical_json({"type": "node_db_candidates", "req_id": req_id, "owner": owner, "items": items}))
        except Exception:
            pass

    async def _handle_db_has_newer(self, ws, msg: dict) -> None:
        req_id = msg.get("req_id")
        owner = msg.get("owner")
        since = msg.get("since_created_at")
        if not isinstance(req_id, str) or not isinstance(owner, str):
            return
        try:
            since_i = int(since)
        except Exception:
            since_i = 0
        items = self._list_db_items(owner, limit=1, since_created_at=since_i)
        item = items[0] if items else None
        try:
            await ws.send(canonical_json({"type": "node_db_has_newer", "req_id": req_id, "owner": owner, "item": item}))
        except Exception:
            pass

    async def _handle_fetch(self, ws, msg: dict):
        req_id = msg.get("req_id")
        blob_sha = msg.get("blob_sha256")
        if not isinstance(req_id, str) or not isinstance(blob_sha, str):
            return

        self._log(f"fetch req_id={req_id} sha={blob_sha[:12]}..")

        meta = self.store.get_meta(blob_sha) or {}
        if not isinstance(meta, dict):
            meta = {}

        if meta.get("kind") == "file" and not meta.get("file_id"):
            fid2 = self.store.ensure_file_id_for_sha(blob_sha)
            if fid2:
                meta = self.store.get_meta(blob_sha) or meta
        extra = {
            "owner": meta.get("owner"),
            "kind": meta.get("kind"),
            "created_at": meta.get("created_at"),
            "file_id": meta.get("file_id"),
            "sig": (meta.get("sig_b64") if str(meta.get("file_id_source") or "") != "derived" else None),
        }

        blob = self.store.get(blob_sha)
        if blob is None:
            self._log(f"fetch miss req_id={req_id} sha={blob_sha[:12]}..")
            await ws.send(canonical_json({"type": "node_blob", "req_id": req_id, "ok": False, "error": "NOT_FOUND"}))
            return

        MAX_RAW_CHUNK = 20 * 1024 * 1024
        if len(blob) <= MAX_RAW_CHUNK:
            self._log(f"fetch hit(req_id={req_id}) sha={blob_sha[:12]}.. bytes={len(blob)}")
            out = {"type": "node_blob", "req_id": req_id, "ok": True, "blob_sha256": blob_sha, "ct": b64e(blob)}
            out.update(extra)
            await ws.send(canonical_json(out))
            return

        await ws.send(
            canonical_json(
                {
                    "type": "node_blob_begin",
                    "req_id": req_id,
                    "ok": True,
                    "blob_sha256": blob_sha,
                    "bytes_total": int(len(blob)),
                    "owner": extra.get("owner"),
                    "kind": extra.get("kind"),
                    "created_at": extra.get("created_at"),
                    "file_id": extra.get("file_id"),
                    "sig": extra.get("sig"),
                }
            )
        )

        idx = 0
        for off in range(0, len(blob), MAX_RAW_CHUNK):
            chunk = blob[off : off + MAX_RAW_CHUNK]
            if not chunk:
                continue
            await ws.send(canonical_json({"type": "node_blob_chunk", "req_id": req_id, "idx": int(idx), "data": b64e(chunk)}))
            idx += 1

        await ws.send(canonical_json({"type": "node_blob_end", "req_id": req_id}))

    async def _handle_fetch_file(self, ws, msg: dict):
        req_id = msg.get("req_id")
        file_id = msg.get("file_id")
        if not isinstance(req_id, str) or not isinstance(file_id, str):
            return
        if len(file_id) != 64:
            return

        self._log(f"fetch_file req_id={req_id} file_id={file_id[:12]}..")

        sha = self.store.find_file_sha_by_file_id(file_id)
        if not isinstance(sha, str) or not sha:
            self._log(f"fetch_file miss req_id={req_id} file_id={file_id[:12]}..")
            await ws.send(canonical_json({"type": "node_blob", "req_id": req_id, "ok": False, "error": "NOT_FOUND"}))
            return

        meta = self.store.get_meta(sha) or {}
        if not isinstance(meta, dict):
            meta = {}

        if meta.get("kind") == "file" and not meta.get("file_id"):
            fid2 = self.store.ensure_file_id_for_sha(sha)
            if fid2:
                meta = self.store.get_meta(sha) or meta
        extra = {
            "owner": meta.get("owner"),
            "kind": meta.get("kind"),
            "created_at": meta.get("created_at"),
            "file_id": meta.get("file_id"),
            "sig": (meta.get("sig_b64") if str(meta.get("file_id_source") or "") != "derived" else None),
        }

        blob = self.store.get(sha)
        if blob is None:
            self._log(f"fetch_file meta_hit_blob_miss req_id={req_id} file_id={file_id[:12]}.. sha={sha[:12]}..")
            await ws.send(canonical_json({"type": "node_blob", "req_id": req_id, "ok": False, "error": "NOT_FOUND"}))
            return

        MAX_RAW_CHUNK = 20 * 1024 * 1024
        if len(blob) <= MAX_RAW_CHUNK:
            self._log(f"fetch_file hit(req_id={req_id}) file_id={file_id[:12]}.. sha={sha[:12]}.. bytes={len(blob)}")
            out = {"type": "node_blob", "req_id": req_id, "ok": True, "blob_sha256": sha, "ct": b64e(blob)}
            out.update(extra)
            await ws.send(canonical_json(out))
            return

        await ws.send(
            canonical_json(
                {
                    "type": "node_blob_begin",
                    "req_id": req_id,
                    "ok": True,
                    "blob_sha256": sha,
                    "bytes_total": int(len(blob)),
                    "owner": extra.get("owner"),
                    "kind": extra.get("kind"),
                    "created_at": extra.get("created_at"),
                    "file_id": extra.get("file_id"),
                    "sig": extra.get("sig"),
                }
            )
        )

        idx = 0
        for off in range(0, len(blob), MAX_RAW_CHUNK):
            chunk = blob[off : off + MAX_RAW_CHUNK]
            if not chunk:
                continue
            await ws.send(canonical_json({"type": "node_blob_chunk", "req_id": req_id, "idx": int(idx), "data": b64e(chunk)}))
            idx += 1

        await ws.send(canonical_json({"type": "node_blob_end", "req_id": req_id}))

def main():
    try:
        sys.stdout.reconfigure(line_buffering=True)
        sys.stderr.reconfigure(line_buffering=True)
    except Exception:
        pass

    parser = argparse.ArgumentParser(description="Antch full-node (cloud blob store)")
    parser.add_argument("--node-config", default=str(APP_DIR / "node.conf"))
    args = parser.parse_args()

    node_cfg_path = ensure_node_config(Path(args.node_config))
    cfg = NodeConfig.from_file(node_cfg_path)
    store = BlobStore(cfg.data_dir, cfg.max_storage_mb)
    client = NodeClient(cfg, store)

    asyncio.run(client.run_forever())

if __name__ == "__main__":
    main()






























from .base import (

    Plugin,
    PluginMeta,
    PluginSetting,
    PluginSettingType,
    HookType,


    RegisteredButton,
    RegisteredPanel,
    RegisteredContextMenuItem,
    RegisteredKeybinding,
    RegisteredCommand,
    RegisteredWidget,
    RegisteredSettingsTab,


    PluginError,
    PluginLoadError,
    PluginSettingsError,
    PluginDependencyError,
)

from .loader import PluginLoader, get_plugin_loader

__all__ = [

    "Plugin",
    "PluginMeta",
    "PluginSetting",
    "PluginSettingType",
    "HookType",


    "RegisteredButton",
    "RegisteredPanel",
    "RegisteredContextMenuItem",
    "RegisteredKeybinding",
    "RegisteredCommand",
    "RegisteredWidget",
    "RegisteredSettingsTab",


    "PluginLoader",
    "get_plugin_loader",


    "PluginError",
    "PluginLoadError",
    "PluginSettingsError",
    "PluginDependencyError",
]



from __future__ import annotations

from textual import events
from textual.message import Message
from textual.widgets import Button, Static, ListView, ListItem
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll

class CallParticipantItem(Static):

    class ContextRequested(Message):
        def __init__(self, sender: "CallParticipantItem", alias: str, x: int, y: int):
            super().__init__()
            self.sender = sender
            self.alias = alias
            self.x = x
            self.y = y

    def __init__(self, alias: str, is_speaking: bool = False, is_muted: bool = False):
        super().__init__(alias, classes="call-participant")
        self.alias = alias
        self._is_speaking = is_speaking
        self._is_muted = is_muted
        self._update_display()

    def _update_display(self) -> None:
        icon = "🔇" if self._is_muted else ("🎤" if self._is_speaking else "👤")
        self.update(f"{icon} {self.alias}")

    def set_speaking(self, speaking: bool) -> None:
        self._is_speaking = speaking
        self._update_display()

    def set_muted(self, muted: bool) -> None:
        self._is_muted = muted
        self._update_display()

    def set_label(self, text: str) -> None:
        self.update(text)

    def on_mouse_down(self, event: events.MouseDown) -> None:
        if event.button == 3:
            self.post_message(self.ContextRequested(self, self.alias, event.screen_x, event.screen_y))
            event.stop()

class GroupMemberItem(Static):

    class ContextRequested(Message):
        def __init__(self, alias: str, x: int, y: int):
            super().__init__()
            self.alias = alias
            self.x = x
            self.y = y

    def __init__(self, alias: str, online: bool = False, in_call: bool = False):
        super().__init__("", classes="gmp-member")
        self.alias = alias
        self._online = online
        self._in_call = in_call
        self._update_display()

    def _update_display(self) -> None:
        if self._in_call:
            icon = "🎤"
            self.add_class("in-call")
            self.remove_class("online")
        elif self._online:
            icon = "🟢"
            self.add_class("online")
            self.remove_class("in-call")
        else:
            icon = "⚫"
            self.remove_class("online")
            self.remove_class("in-call")
        self.update(f"{icon} {self.alias}")

    def on_mouse_down(self, event: events.MouseDown) -> None:
        if event.button == 3:
            self.post_message(self.ContextRequested(self.alias, event.screen_x, event.screen_y))
            event.stop()

class GroupMembersPanel(Vertical):

    DEFAULT_CSS = """
    GroupMembersPanel {
        width: 0;
        min-width: 0;
        max-width: 0;
        height: 100%;
        dock: right;
        background: $surface;
        border-left: solid $surface-darken-1;
        display: none;
        overflow: hidden;
    }
    GroupMembersPanel.visible {
        width: 18;
        min-width: 18;
        max-width: 18;
        display: block;
    }
    GroupMembersPanel #gmp_header {
        height: 1;
        width: 100%;
        padding: 0 1;
        background: $surface-darken-1;
    }
    GroupMembersPanel #gmp_title {
        width: 1fr;
        height: 1;
        text-align: left;
        color: $text-muted;
    }
    GroupMembersPanel #gmp_close {
        width: 1;
        height: 1;
        min-width: 1;
        padding: 0;
        border: none;
        background: transparent;
        color: $text-muted;
    }
    GroupMembersPanel #gmp_close:hover {
        color: $error;
    }
    GroupMembersPanel #gmp_list {
        height: 1fr;
        width: 100%;
        padding: 0;
        overflow-y: auto;
    }
    GroupMembersPanel .gmp-member {
        height: 1;
        width: 100%;
        padding: 0 1;
        background: transparent;
    }
    GroupMembersPanel .gmp-member:hover {
        background: $surface-lighten-1;
    }
    GroupMembersPanel .gmp-member.online {
        color: $success;
    }
    GroupMembersPanel .gmp-member.in-call {
        color: $warning;
    }
    """

    class CloseRequested(Message):
        def __init__(self, sender: "GroupMembersPanel"):
            super().__init__()
            self.sender = sender

    class CallRequested(Message):
        def __init__(self, sender: "GroupMembersPanel", group_id: str):
            super().__init__()
            self.sender = sender
            self.group_id = group_id

    class MemberContextRequested(Message):
        def __init__(self, alias: str, group_id: str, x: int, y: int):
            super().__init__()
            self.alias = alias
            self.group_id = group_id
            self.x = x
            self.y = y

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._group_id: str = ""
        self._group_name: str = "Group"
        self._members: list[dict] = []
        self._active_call_participants: int = 0

    def compose(self) -> ComposeResult:
        with Horizontal(id="gmp_header"):
            yield Static("Members", id="gmp_title")
            yield Button("×", id="gmp_close")
        with VerticalScroll(id="gmp_list"):
            pass

    def show(self, group_id: str, group_name: str = "Group") -> None:
        self._group_id = group_id
        self._group_name = group_name
        self.add_class("visible")
        self._update_title()

    def hide(self) -> None:
        self.remove_class("visible")

    def toggle(self, group_id: str, group_name: str = "Group") -> None:
        if self.has_class("visible") and self._group_id == group_id:
            self.hide()
        else:
            self.show(group_id, group_name)

    def _update_title(self) -> None:
        try:
            title = self.query_one("#gmp_title", Static)
            online_count = sum(1 for m in self._members if m.get("online"))
            title.update(f"{online_count}/{len(self._members)}")
        except Exception:
            pass

    def update_members(self, members: list[dict]) -> None:
        self._members = list(members)
        self._update_title()
        self._rebuild_list()

    def set_active_call(self, participant_count: int) -> None:
        self._active_call_participants = participant_count

    def _rebuild_list(self) -> None:
        try:
            container = self.query_one("#gmp_list", VerticalScroll)
            container.remove_children()

            sorted_members = sorted(
                self._members,
                key=lambda m: (not m.get("in_call", False), not m.get("online", False), m.get("alias", "").lower())
            )
            for member in sorted_members:
                alias = member.get("alias", "")
                online = member.get("online", False)
                in_call = member.get("in_call", False)

                item = GroupMemberItem(alias, online=online, in_call=in_call)
                container.mount(item)
        except Exception:
            pass

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "gmp_close":
            self.post_message(self.CloseRequested(self))
            event.stop()

    def on_group_member_item_context_requested(self, event: GroupMemberItem.ContextRequested) -> None:
        self.post_message(self.MemberContextRequested(event.alias, self._group_id, event.x, event.y))
        event.stop()

    def on_mouse_down(self, event: events.MouseDown) -> None:
        if event.button == 3:

            event.stop()

class CallVolumeBar(Static):

    class Changed(Message):
        def __init__(self, sender: "CallVolumeBar", gain: float):
            super().__init__()
            self.sender = sender
            self.gain = float(gain)

    def __init__(self, gain: float = 1.0, **kwargs):
        super().__init__("", **kwargs)
        self._gain = max(0.0, min(3.0, float(gain)))
        self._dragging = False
        self._update_bar()

    @property
    def gain(self) -> float:
        return float(self._gain)

    def set_gain(self, gain: float) -> None:
        self._gain = max(0.0, min(3.0, float(gain)))
        self._update_bar()

    def _update_bar(self) -> None:
        try:
            w = max(10, int(self.size.width) - 5)
        except:
            w = 20

        pct = self._gain / 3.0
        fill = int(pct * w)
        bar = "█" * fill + "─" * (w - fill)
        self.update(f"{int(pct * 100):>3}% {bar}")

    def _gain_from_x(self, local_x: int) -> float:
        try:
            prefix = 5
            bar_x = local_x - prefix
            bar_w = max(1, self.size.width - prefix)
            ratio = max(0.0, min(1.0, bar_x / bar_w))
            return ratio * 3.0
        except:
            return self._gain

    def on_mouse_down(self, event: events.MouseDown) -> None:
        if event.button == 1:
            self._dragging = True
            self.capture_mouse()
            self._gain = self._gain_from_x(event.x)
            self._update_bar()
            self.post_message(self.Changed(self, self._gain))
            event.stop()

    def on_mouse_move(self, event: events.MouseMove) -> None:
        if self._dragging:
            new_gain = self._gain_from_x(event.x)
            if abs(new_gain - self._gain) > 0.01:
                self._gain = new_gain
                self._update_bar()
                self.post_message(self.Changed(self, self._gain))

    def on_mouse_up(self, event: events.MouseUp) -> None:
        if self._dragging:
            self._dragging = False
            self.release_mouse()

class CallUserPopover(Vertical):

    class Closed(Message):
        def __init__(self, sender: "CallUserPopover"):
            super().__init__()
            self.sender = sender

    class RemoteMuteToggled(Message):
        def __init__(self, sender: "CallUserPopover", alias: str):
            super().__init__()
            self.sender = sender
            self.alias = alias

    class SelfMuteToggled(Message):
        def __init__(self, sender: "CallUserPopover"):
            super().__init__()
            self.sender = sender

    class SelfDeafenToggled(Message):
        def __init__(self, sender: "CallUserPopover"):
            super().__init__()
            self.sender = sender

    class RemoteGainChanged(Message):
        def __init__(self, sender: "CallUserPopover", alias: str, gain: float):
            super().__init__()
            self.sender = sender
            self.alias = alias
            self.gain = float(gain)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._alias = ""
        self._is_self = False
        self._remote_muted = False
        self._self_muted = False
        self._self_deafened = False
        self._gain = 1.0

    def compose(self) -> ComposeResult:

        with Horizontal(id="pop_header"):
            yield Static("", id="pop_title")
            yield Button("X", id="pop_close")

        with Vertical(id="pop_body"):

            yield Button("MUTE USER", id="pop_remote_mute")
            with Horizontal(id="pop_vol_row"):
                yield Static("VOL:", id="pop_vol_label")
                yield CallVolumeBar(id="pop_vol")

            yield Button("MUTE", id="pop_self_mute")
            yield Button("DEAFEN", id="pop_self_deafen")

    def open(self, *, alias: str, x: int, y: int, is_self: bool,
             remote_muted: bool, self_muted: bool, self_deafened: bool, gain: float) -> None:
        self._alias = alias
        self._is_self = is_self
        self._remote_muted = remote_muted
        self._self_muted = self_muted
        self._self_deafened = self_deafened
        self._gain = gain

        W, H = 50, 12
        sw, sh = self.app.size.width, self.app.size.height

        parent_x, parent_y = 0, 0
        try:
            parent = self.parent
            if parent is not None and hasattr(parent, "region"):
                parent_x = parent.region.x
                parent_y = parent.region.y
        except Exception:
            pass

        px = x - parent_x
        py = y - parent_y

        px = max(0, min(px, sw - W - 2))
        py = max(0, min(py, sh - H - 2))

        self.styles.display = "block"
        self.styles.visibility = "visible"
        self.styles.width = W
        self.styles.height = H
        self.styles.offset = (px, py)
        self._update_ui()
        try:
            self.refresh(layout=True, repaint=True)
        except Exception:
            pass

    def close(self) -> None:
        self.styles.display = "none"
        self.styles.visibility = "hidden"
        self.styles.width = 0
        self.styles.height = 0
        self.styles.offset = (-9999, -9999)
        try:
            self.refresh(layout=True, repaint=True)
        except Exception:
            pass

    def _update_ui(self) -> None:
        try:
            self.query_one("#pop_title", Static).update(f" {self._alias}")
        except:
            pass

        try:
            self.query_one("#pop_remote_mute").styles.display = "none" if self._is_self else "block"
            self.query_one("#pop_vol_row").styles.display = "none" if self._is_self else "block"
            self.query_one("#pop_self_mute").styles.display = "block" if self._is_self else "none"
            self.query_one("#pop_self_deafen").styles.display = "block" if self._is_self else "none"
        except:
            pass

        try:
            if self._is_self:
                m = self.query_one("#pop_self_mute", Button)
                d = self.query_one("#pop_self_deafen", Button)
                m.label = "🔇 MUTE: ON" if self._self_muted else "🎤 MUTE: OFF"
                d.label = "🔇 DEAFEN: ON" if self._self_deafened else "🔊 DEAFEN: OFF"
                m.set_class(self._self_muted, "active")
                d.set_class(self._self_deafened, "active")
            else:
                r = self.query_one("#pop_remote_mute", Button)
                r.label = "🔇 MUTED" if self._remote_muted else "🔊 UNMUTED"
                r.set_class(self._remote_muted, "active")
                self.query_one("#pop_vol", CallVolumeBar).set_gain(self._gain)
        except:
            pass

    def on_call_volume_bar_changed(self, event: CallVolumeBar.Changed) -> None:
        if not self._is_self:
            self._gain = event.gain
            self.post_message(self.RemoteGainChanged(self, self._alias, event.gain))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        bid = event.button.id
        if bid == "pop_close":
            self.post_message(self.Closed(self))
        elif bid == "pop_remote_mute" and not self._is_self:
            self.post_message(self.RemoteMuteToggled(self, self._alias))
        elif bid == "pop_self_mute" and self._is_self:
            self.post_message(self.SelfMuteToggled(self))
        elif bid == "pop_self_deafen" and self._is_self:
            self.post_message(self.SelfDeafenToggled(self))
        event.stop()

    def on_key(self, event: events.Key) -> None:
        if event.key == "escape":
            self.post_message(self.Closed(self))
            event.stop()



from __future__ import annotations

THEMES: dict[str, dict[str, str]] = {
    "default": {
        "bg": "#121212",
        "panel": "#1e1e1e",
        "text": "#e0e0e0",
        "muted": "#808080",
        "border": "#333333",
        "accent": "#6b5b7a",
        "accent_text": "#e0e0e0",
        "hover": "#2c2c2c",
    },
    "gruvbox": {
        "bg": "#fbf1c7",
        "panel": "#ebdbb2",
        "text": "#3c3836",
        "muted": "#665c54",
        "border": "#d5c4a1",
        "accent": "#665c54",
        "accent_text": "#fbf1c7",
        "hover": "#f2e5bc",
    },
    "gruvbox-dark": {
        "bg": "#282828",
        "panel": "#32302f",
        "text": "#ebdbb2",
        "muted": "#a89984",
        "border": "#504945",
        "accent": "#665c54",
        "accent_text": "#ebdbb2",
        "hover": "#3c3836",
    },
    "dracula": {
        "bg": "#282a36",
        "panel": "#21222c",
        "text": "#f8f8f2",
        "muted": "#b3b3b3",
        "border": "#44475a",
        "accent": "#6272a4",
        "accent_text": "#f8f8f2",
        "hover": "#2f3140",
    },
    "nord": {
        "bg": "#2e3440",
        "panel": "#3b4252",
        "text": "#eceff4",
        "muted": "#d8dee9",
        "border": "#4c566a",
        "accent": "#5e6779",
        "accent_text": "#eceff4",
        "hover": "#434c5e",
    },
    "solarized-light": {
        "bg": "#fdf6e3",
        "panel": "#eee8d5",
        "text": "#586e75",
        "muted": "#657b83",
        "border": "#ddd6c1",
        "accent": "#657b83",
        "accent_text": "#fdf6e3",
        "hover": "#f5efdc",
    },
    "solarized-dark": {
        "bg": "#002b36",
        "panel": "#073642",
        "text": "#eee8d5",
        "muted": "#93a1a1",
        "border": "#0f4a5a",
        "accent": "#586e75",
        "accent_text": "#eee8d5",
        "hover": "#0a3f4d",
    },
    "catppuccin-mocha": {
        "bg": "#1e1e2e",
        "panel": "#181825",
        "text": "#cdd6f4",
        "muted": "#a6adc8",
        "border": "#313244",
        "accent": "#585b70",
        "accent_text": "#cdd6f4",
        "hover": "#242438",
    },

    "catppuccin-latte": {
        "bg": "#eff1f5",
        "panel": "#e6e9ef",
        "text": "#4c4f69",
        "muted": "#6c6f85",
        "border": "#ccd0da",
        "accent": "#7c7f93",
        "accent_text": "#eff1f5",
        "hover": "#dce0e8",
    },
    "catppuccin-frappe": {
        "bg": "#303446",
        "panel": "#292c3c",
        "text": "#c6d0f5",
        "muted": "#a5adce",
        "border": "#414559",
        "accent": "#626880",
        "accent_text": "#c6d0f5",
        "hover": "#3b3f52",
    },
    "catppuccin-macchiato": {
        "bg": "#24273a",
        "panel": "#1e2030",
        "text": "#cad3f5",
        "muted": "#a5adcb",
        "border": "#363a4f",
        "accent": "#5b6078",
        "accent_text": "#cad3f5",
        "hover": "#2e3244",
    },
    "tokyo-night": {
        "bg": "#1a1b26",
        "panel": "#16161e",
        "text": "#c0caf5",
        "muted": "#565f89",
        "border": "#292e42",
        "accent": "#565f89",
        "accent_text": "#c0caf5",
        "hover": "#24283b",
    },
    "tokyo-night-storm": {
        "bg": "#24283b",
        "panel": "#1f2335",
        "text": "#c0caf5",
        "muted": "#565f89",
        "border": "#3b4261",
        "accent": "#565f89",
        "accent_text": "#c0caf5",
        "hover": "#292e42",
    },
    "tokyo-night-light": {
        "bg": "#d5d6db",
        "panel": "#cbccd1",
        "text": "#343b58",
        "muted": "#6172af",
        "border": "#b4b5ba",
        "accent": "#6e7191",
        "accent_text": "#d5d6db",
        "hover": "#c4c5ca",
    },
    "one-dark": {
        "bg": "#282c34",
        "panel": "#21252b",
        "text": "#abb2bf",
        "muted": "#5c6370",
        "border": "#3e4451",
        "accent": "#5c6370",
        "accent_text": "#abb2bf",
        "hover": "#2c323c",
    },
    "one-light": {
        "bg": "#fafafa",
        "panel": "#f0f0f0",
        "text": "#383a42",
        "muted": "#a0a1a7",
        "border": "#d3d3d3",
        "accent": "#8a8b91",
        "accent_text": "#383a42",
        "hover": "#e5e5e5",
    },
    "monokai-pro": {
        "bg": "#2d2a2e",
        "panel": "#221f22",
        "text": "#fcfcfa",
        "muted": "#939293",
        "border": "#403e41",
        "accent": "#5b5759",
        "accent_text": "#fcfcfa",
        "hover": "#363337",
    },
    "github-dark": {
        "bg": "#0d1117",
        "panel": "#161b22",
        "text": "#c9d1d9",
        "muted": "#8b949e",
        "border": "#30363d",
        "accent": "#484f58",
        "accent_text": "#c9d1d9",
        "hover": "#21262d",
    },
    "github-light": {
        "bg": "#ffffff",
        "panel": "#f6f8fa",
        "text": "#24292f",
        "muted": "#57606a",
        "border": "#d0d7de",
        "accent": "#6e7781",
        "accent_text": "#24292f",
        "hover": "#eaeef2",
    },
    "material-darker": {
        "bg": "#212121",
        "panel": "#1a1a1a",
        "text": "#eeffff",
        "muted": "#545454",
        "border": "#303030",
        "accent": "#4a4a4a",
        "accent_text": "#eeffff",
        "hover": "#2a2a2a",
    },
    "material-palenight": {
        "bg": "#292d3e",
        "panel": "#232635",
        "text": "#a6accd",
        "muted": "#676e95",
        "border": "#3a3f58",
        "accent": "#575d78",
        "accent_text": "#a6accd",
        "hover": "#32374a",
    },
    "ayu-dark": {
        "bg": "#0b0e14",
        "panel": "#0d1017",
        "text": "#bfbdb6",
        "muted": "#565b66",
        "border": "#1c2029",
        "accent": "#464b55",
        "accent_text": "#bfbdb6",
        "hover": "#151920",
    },
    "ayu-mirage": {
        "bg": "#1f2430",
        "panel": "#1a1f29",
        "text": "#cbccc6",
        "muted": "#707a8c",
        "border": "#2d3444",
        "accent": "#5c6773",
        "accent_text": "#cbccc6",
        "hover": "#272d3a",
    },
    "ayu-light": {
        "bg": "#fafafa",
        "panel": "#f3f4f5",
        "text": "#575f66",
        "muted": "#828c99",
        "border": "#e7e8e9",
        "accent": "#8a9199",
        "accent_text": "#575f66",
        "hover": "#eeeff0",
    },
    "rose-pine": {
        "bg": "#191724",
        "panel": "#1f1d2e",
        "text": "#e0def4",
        "muted": "#6e6a86",
        "border": "#26233a",
        "accent": "#524f67",
        "accent_text": "#e0def4",
        "hover": "#21202e",
    },
    "rose-pine-moon": {
        "bg": "#232136",
        "panel": "#2a273f",
        "text": "#e0def4",
        "muted": "#6e6a86",
        "border": "#393552",
        "accent": "#56526e",
        "accent_text": "#e0def4",
        "hover": "#2d2a41",
    },
    "rose-pine-dawn": {
        "bg": "#faf4ed",
        "panel": "#f2e9e1",
        "text": "#575279",
        "muted": "#9893a5",
        "border": "#dfdad9",
        "accent": "#797593",
        "accent_text": "#faf4ed",
        "hover": "#f4ede8",
    },
    "everforest-dark": {
        "bg": "#2d353b",
        "panel": "#272e33",
        "text": "#d3c6aa",
        "muted": "#859289",
        "border": "#3d484d",
        "accent": "#5c6a5e",
        "accent_text": "#d3c6aa",
        "hover": "#343e44",
    },
    "everforest-light": {
        "bg": "#fdf6e3",
        "panel": "#f4f0d9",
        "text": "#5c6a72",
        "muted": "#829181",
        "border": "#e0dcc7",
        "accent": "#708374",
        "accent_text": "#5c6a72",
        "hover": "#f0ead4",
    },
    "kanagawa": {
        "bg": "#1f1f28",
        "panel": "#16161d",
        "text": "#dcd7ba",
        "muted": "#727169",
        "border": "#2a2a37",
        "accent": "#54546d",
        "accent_text": "#dcd7ba",
        "hover": "#252530",
    },
    "kanagawa-light": {
        "bg": "#f2ecbc",
        "panel": "#e7dba0",
        "text": "#545464",
        "muted": "#8a8980",
        "border": "#d5cea6",
        "accent": "#71717a",
        "accent_text": "#f2ecbc",
        "hover": "#e8e0a9",
    },
    "night-owl": {
        "bg": "#011627",
        "panel": "#001122",
        "text": "#d6deeb",
        "muted": "#637777",
        "border": "#122d42",
        "accent": "#456578",
        "accent_text": "#d6deeb",
        "hover": "#0b2942",
    },
    "synthwave-84": {
        "bg": "#262335",
        "panel": "#1e1a2b",
        "text": "#ffffff",
        "muted": "#848bbd",
        "border": "#34294f",
        "accent": "#524668",
        "accent_text": "#ffffff",
        "hover": "#2e2943",
    },
    "cyberpunk": {
        "bg": "#0a0a0f",
        "panel": "#12121a",
        "text": "#d0d0d0",
        "muted": "#606060",
        "border": "#1a1a25",
        "accent": "#3a3a45",
        "accent_text": "#d0d0d0",
        "hover": "#15151f",
    },
    "high-contrast": {
        "bg": "#000000",
        "panel": "#0a0a0a",
        "text": "#ffffff",
        "muted": "#aaaaaa",
        "border": "#333333",
        "accent": "#555555",
        "accent_text": "#ffffff",
        "hover": "#1a1a1a",
    },
    "paper": {
        "bg": "#f5f5dc",
        "panel": "#ebe7d3",
        "text": "#3b3228",
        "muted": "#6b5e4a",
        "border": "#d4cfbb",
        "accent": "#6b5e4a",
        "accent_text": "#f5f5dc",
        "hover": "#e6e1cc",
    },
}

def _parse_kde_color(color_str: str) -> str | None:
    if not color_str:
        return None
    try:

        parts = [p.strip() for p in color_str.strip().split(",")]
        if len(parts) >= 3:
            r, g, b = int(parts[0]), int(parts[1]), int(parts[2])
            return f"#{r:02x}{g:02x}{b:02x}"
    except Exception:
        pass
    return None

def _adjust_color(hex_color: str, amount: int) -> str:
    try:
        hex_color = hex_color.lstrip("#")
        r = max(0, min(255, int(hex_color[0:2], 16) + amount))
        g = max(0, min(255, int(hex_color[2:4], 16) + amount))
        b = max(0, min(255, int(hex_color[4:6], 16) + amount))
        return f"#{r:02x}{g:02x}{b:02x}"
    except Exception:
        return hex_color

def _get_luminance(hex_color: str) -> float:
    try:
        hex_color = hex_color.lstrip("#")
        r, g, b = int(hex_color[0:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)
        return (r * 299 + g * 587 + b * 114) / 1000
    except Exception:
        return 128

def _read_kde_colors_file(filepath: str) -> dict[str, str] | None:
    import os

    if not os.path.exists(filepath):
        return None

    try:
        with open(filepath, "r") as f:
            content = f.read()

        colors = {}
        current_section = None

        for line in content.split("\n"):
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            if line.startswith("[") and line.endswith("]"):
                current_section = line[1:-1]
                continue

            if "=" in line:
                key, _, value = line.partition("=")
                key = key.strip()
                value = value.strip()

                if current_section == "Colors:Window":
                    if key == "BackgroundNormal":
                        colors["bg"] = _parse_kde_color(value)
                    elif key == "ForegroundNormal":
                        colors["text"] = _parse_kde_color(value)
                    elif key == "ForegroundInactive":
                        colors["muted"] = _parse_kde_color(value)

                elif current_section == "Colors:View":
                    if key == "BackgroundNormal":
                        colors["panel"] = _parse_kde_color(value)
                    elif key == "BackgroundAlternate":
                        colors["hover"] = _parse_kde_color(value)

                elif current_section == "Colors:Button":
                    if key == "BackgroundNormal" and not colors.get("hover"):
                        colors["hover"] = _parse_kde_color(value)
                    elif key == "DecorationFocus":
                        colors["accent"] = _parse_kde_color(value)

                elif current_section == "Colors:Selection":
                    if key == "BackgroundNormal":
                        colors["accent"] = _parse_kde_color(value)
                    elif key == "ForegroundNormal":
                        colors["accent_text"] = _parse_kde_color(value)

                elif current_section == "WM":
                    if key == "activeBackground":
                        colors["border"] = _parse_kde_color(value)
                    elif key == "inactiveBackground" and not colors.get("border"):
                        colors["border"] = _parse_kde_color(value)

        return colors if colors else None
    except Exception:
        return None

def _complete_theme_colors(colors: dict[str, str]) -> dict[str, str] | None:
    if not colors:
        return None

    bg = colors.get("bg")
    text = colors.get("text")

    if not bg or not text:
        return None

    is_dark = _get_luminance(bg) < 128

    if not colors.get("panel"):

        colors["panel"] = _adjust_color(bg, 15 if is_dark else -10)

    if not colors.get("muted"):

        if is_dark:
            colors["muted"] = _adjust_color(text, -60)
        else:
            colors["muted"] = _adjust_color(text, 60)

    if not colors.get("border"):

        colors["border"] = _adjust_color(bg, 40 if is_dark else -30)

    if not colors.get("accent"):

        colors["accent"] = "#5294e2" if is_dark else "#3584e4"

    if not colors.get("accent_text"):

        accent_lum = _get_luminance(colors["accent"])
        colors["accent_text"] = "#ffffff" if accent_lum < 128 else "#000000"

    if not colors.get("hover"):

        colors["hover"] = _adjust_color(bg, 20 if is_dark else -15)

    colors["is_dark"] = is_dark
    return colors

def _read_kdeglobals() -> dict[str, str] | None:
    import os

    colors = _read_kde_colors_file(os.path.expanduser("~/.config/kdeglobals"))
    if colors:
        return _complete_theme_colors(colors)

    return None

def _read_kde_colorscheme() -> dict[str, str] | None:
    import os
    import subprocess

    scheme_name = None

    for cmd in ["kreadconfig5", "kreadconfig6"]:
        try:
            result = subprocess.run(
                [cmd, "--file", "kdeglobals", "--group", "General", "--key", "ColorScheme"],
                capture_output=True, text=True, timeout=2
            )
            if result.returncode == 0 and result.stdout.strip():
                scheme_name = result.stdout.strip()
                break
        except Exception:
            continue

    if not scheme_name:
        return None

    search_paths = [
        os.path.expanduser(f"~/.local/share/color-schemes/{scheme_name}.colors"),
        f"/usr/share/color-schemes/{scheme_name}.colors",
        os.path.expanduser(f"~/.kde4/share/apps/color-schemes/{scheme_name}.colors"),
        f"/usr/share/kde4/apps/color-schemes/{scheme_name}.colors",
    ]

    for path in search_paths:
        colors = _read_kde_colors_file(path)
        if colors:
            return _complete_theme_colors(colors)

    return None

def _read_gtk_css_colors() -> dict[str, str] | None:
    import os
    import subprocess
    import re

    theme_name = None
    try:
        result = subprocess.run(
            ["gsettings", "get", "org.gnome.desktop.interface", "gtk-theme"],
            capture_output=True, text=True, timeout=2
        )
        if result.returncode == 0:
            theme_name = result.stdout.strip().strip("'\"")
    except Exception:
        pass

    if not theme_name:
        return None

    search_paths = [
        os.path.expanduser(f"~/.themes/{theme_name}/gtk-3.0/gtk.css"),
        os.path.expanduser(f"~/.local/share/themes/{theme_name}/gtk-3.0/gtk.css"),
        f"/usr/share/themes/{theme_name}/gtk-3.0/gtk.css",
        os.path.expanduser(f"~/.themes/{theme_name}/gtk-4.0/gtk.css"),
        f"/usr/share/themes/{theme_name}/gtk-4.0/gtk.css",
    ]

    css_content = None
    for path in search_paths:
        if os.path.exists(path):
            try:
                with open(path, "r") as f:
                    css_content = f.read()
                break
            except Exception:
                continue

    if not css_content:
        return None

    colors = {}

    def extract_color(value: str) -> str | None:
        if not value:
            return None
        value = value.strip()

        if value.startswith("#"):
            if len(value) == 4:
                return f"#{value[1]*2}{value[2]*2}{value[3]*2}"
            if len(value) >= 7:
                return value[:7]

        rgb_match = re.match(r"rgba?\s*\(\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)", value)
        if rgb_match:
            r, g, b = int(rgb_match.group(1)), int(rgb_match.group(2)), int(rgb_match.group(3))
            return f"#{r:02x}{g:02x}{b:02x}"
        return None

    color_patterns = [

        (r"@define-color\s+theme_bg_color\s+([^;]+);", "bg"),
        (r"@define-color\s+window_bg_color\s+([^;]+);", "bg"),
        (r"@define-color\s+base_color\s+([^;]+);", "bg"),
        (r"@define-color\s+bg_color\s+([^;]+);", "bg"),

        (r"@define-color\s+theme_fg_color\s+([^;]+);", "text"),
        (r"@define-color\s+window_fg_color\s+([^;]+);", "text"),
        (r"@define-color\s+text_color\s+([^;]+);", "text"),
        (r"@define-color\s+fg_color\s+([^;]+);", "text"),

        (r"@define-color\s+theme_base_color\s+([^;]+);", "panel"),
        (r"@define-color\s+view_bg_color\s+([^;]+);", "panel"),
        (r"@define-color\s+card_bg_color\s+([^;]+);", "panel"),

        (r"@define-color\s+theme_selected_bg_color\s+([^;]+);", "accent"),
        (r"@define-color\s+accent_color\s+([^;]+);", "accent"),
        (r"@define-color\s+accent_bg_color\s+([^;]+);", "accent"),
        (r"@define-color\s+selected_bg_color\s+([^;]+);", "accent"),

        (r"@define-color\s+borders\s+([^;]+);", "border"),
        (r"@define-color\s+border_color\s+([^;]+);", "border"),
        (r"@define-color\s+unfocused_borders\s+([^;]+);", "border"),

        (r"@define-color\s+insensitive_fg_color\s+([^;]+);", "muted"),
        (r"@define-color\s+placeholder_text_color\s+([^;]+);", "muted"),
        (r"@define-color\s+dim_label_color\s+([^;]+);", "muted"),

        (r"@define-color\s+theme_unfocused_bg_color\s+([^;]+);", "hover"),
        (r"@define-color\s+hover_bg_color\s+([^;]+);", "hover"),
    ]

    for pattern, key in color_patterns:
        if colors.get(key):
            continue
        match = re.search(pattern, css_content, re.IGNORECASE)
        if match:
            color = extract_color(match.group(1))
            if color:
                colors[key] = color

    return _complete_theme_colors(colors)

def _read_gtk_dconf_colors() -> dict[str, str] | None:
    import subprocess

    colors = {}

    try:
        result = subprocess.run(
            ["gsettings", "get", "org.gnome.desktop.interface", "accent-color"],
            capture_output=True, text=True, timeout=2
        )
        if result.returncode == 0 and result.stdout.strip():
            accent_name = result.stdout.strip().strip("'\"").lower()

            accent_map = {
                "blue": "#3584e4",
                "teal": "#2190a4",
                "green": "#3a944a",
                "yellow": "#c88800",
                "orange": "#ed5b00",
                "red": "#e62d42",
                "pink": "#d56199",
                "purple": "#9141ac",
                "slate": "#6f8396",
            }
            if accent_name in accent_map:
                colors["accent"] = accent_map[accent_name]
    except Exception:
        pass

    is_dark = None
    try:
        result = subprocess.run(
            ["gsettings", "get", "org.gnome.desktop.interface", "color-scheme"],
            capture_output=True, text=True, timeout=2
        )
        if result.returncode == 0:
            is_dark = "dark" in result.stdout.lower()
    except Exception:
        pass

    if is_dark is None:
        try:
            result = subprocess.run(
                ["gsettings", "get", "org.gnome.desktop.interface", "gtk-theme"],
                capture_output=True, text=True, timeout=2
            )
            if result.returncode == 0:
                is_dark = "dark" in result.stdout.lower()
        except Exception:
            pass

    if is_dark is not None:

        if is_dark:
            colors.setdefault("bg", "#242424")
            colors.setdefault("panel", "#303030")
            colors.setdefault("text", "#ffffff")
            colors.setdefault("muted", "#9a9996")
            colors.setdefault("border", "#1e1e1e")
            colors.setdefault("hover", "#383838")
            colors.setdefault("accent", colors.get("accent", "#3584e4"))
            colors.setdefault("accent_text", "#ffffff")
        else:
            colors.setdefault("bg", "#fafafa")
            colors.setdefault("panel", "#ffffff")
            colors.setdefault("text", "#000000")
            colors.setdefault("muted", "#77767b")
            colors.setdefault("border", "#deddda")
            colors.setdefault("hover", "#f0f0f0")
            colors.setdefault("accent", colors.get("accent", "#3584e4"))
            colors.setdefault("accent_text", "#ffffff")

        colors["is_dark"] = is_dark
        return colors

    return None

def get_gtk_theme_colors() -> dict[str, str] | None:

    colors = _read_gtk_css_colors()
    if colors:
        return colors

    colors = _read_gtk_dconf_colors()
    if colors:
        return colors

    return None

def get_qt_theme_colors() -> dict[str, str] | None:

    colors = _read_kde_colorscheme()
    if colors:
        return colors

    colors = _read_kdeglobals()
    if colors:
        return colors

    return None

def normalize_theme(name: str | None) -> str:
    n = (name or "default").strip().replace("_", "-")
    return n if n in THEMES else "default"

def theme_list() -> list[str]:
    return list(THEMES.keys())

def css_for_theme(name: str, c: dict[str, str]) -> str:
    sel = f"App.theme-{name}"
    bg = c["bg"]
    panel = c["panel"]
    text = c["text"]
    muted = c["muted"]
    border = c["border"]
    accent = c["accent"]
    accent_text = c["accent_text"]
    hover = c["hover"]

    def blend_color(c1: str, c2: str, ratio: float = 0.15) -> str:
        try:
            c1 = c1.lstrip('#')
            c2 = c2.lstrip('#')
            r1, g1, b1 = int(c1[0:2], 16), int(c1[2:4], 16), int(c1[4:6], 16)
            r2, g2, b2 = int(c2[0:2], 16), int(c2[2:4], 16), int(c2[4:6], 16)
            r = int(r1 * (1 - ratio) + r2 * ratio)
            g = int(g1 * (1 - ratio) + g2 * ratio)
            b = int(b1 * (1 - ratio) + b2 * ratio)
            return f"#{r:02x}{g:02x}{b:02x}"
        except:
            return c1

    def lighten(hex_color: str, amount: int = 20) -> str:
        try:
            hex_color = hex_color.lstrip('#')
            r = min(255, int(hex_color[0:2], 16) + amount)
            g = min(255, int(hex_color[2:4], 16) + amount)
            b = min(255, int(hex_color[4:6], 16) + amount)
            return f"#{r:02x}{g:02x}{b:02x}"
        except:
            return hex_color

    def darken(hex_color: str, amount: int = 20) -> str:
        try:
            hex_color = hex_color.lstrip('#')
            r = max(0, int(hex_color[0:2], 16) - amount)
            g = max(0, int(hex_color[2:4], 16) - amount)
            b = max(0, int(hex_color[4:6], 16) - amount)
            return f"#{r:02x}{g:02x}{b:02x}"
        except:
            return hex_color

    accent_soft = blend_color(accent, panel, 0.3)
    accent_softer = blend_color(accent, panel, 0.2)

    input_bg = lighten(bg, 3)

    bg_lighten1 = lighten(bg, 15)
    bg_lighten2 = lighten(bg, 30)
    bg_darken1 = darken(bg, 15)
    bg_darken2 = darken(bg, 30)
    panel_lighten1 = lighten(panel, 15)
    panel_darken1 = darken(panel, 15)
    accent_lighten1 = lighten(accent, 20)
    accent_darken1 = darken(accent, 20)
    border_light = lighten(border, 20)
    border_dark = darken(border, 20)

    return f"""
/* ===== Theme: {name} ===== */
{sel} Screen {{ background: {bg}; color: {text}; }}
{sel} #root, {sel} #main, {sel} #messages, {sel} #settings_content {{ background: {bg}; }}
{sel} #sidebar_full, {sel} #sidebar_rail, {sel} #bottom_bar, {sel} #settings_nav {{ background: {panel}; }}
{sel} #composer_frame {{ background: {bg}; }}
{sel} #sidebar_full, {sel} #sidebar_rail, {sel} #settings_nav {{ border-right: solid {border}; }}
{sel} #bottom_bar, {sel} #composer_frame {{ border-top: solid {border}; }}
{sel} #system_log {{ background: {hover}; color: {muted}; border-bottom: solid {border}; }}
{sel} #info_tab {{ background: {panel}; color: {text}; }}
{sel} #sidebar_handle {{ background: {border}; }}
{sel} #top_info_panel {{ background: {bg}; }}

/* Tabs */
{sel} TabbedContent, {sel} TabPane, {sel} TabBar {{ background: {panel}; border: none; }}
{sel} Tab {{ background: {panel}; color: {muted}; }}
{sel} Tab.-active {{ background: {bg}; color: {text}; }}
{sel} Tab:hover {{ background: {hover}; }}

/* Lists */
{sel} ListView {{ background: {bg}; }}
{sel} ListItem {{ background: transparent; }}
{sel} ListItem:hover {{ background: {hover}; }}
{sel} ListItem.-highlight {{ background: {hover}; border-left: solid {border}; }}
{sel} .chat_item_title, {sel} .req_item_title {{ color: {text}; }}
{sel} .chat_item_sub, {sel} .req_item_sub {{ color: {muted}; }}

/* Buttons - subtle background difference */
{sel} Button {{ background: {hover}; color: {text}; }}
{sel} #btn_settings {{ background: {hover}; color: {text}; }}
{sel} Button:hover {{ background: {bg_lighten1}; }}
{sel} Button.-primary {{ background: {bg_lighten1}; color: {text}; }}
{sel} Button.-primary:hover {{ background: {bg_lighten2}; }}
{sel} Button:focus {{ background: {bg_lighten1}; }}

/* Flat button style (default) - removes 3D borders */
{sel}.btn-style-flat Button {{ border: none; border-top: none; border-bottom: none; }}

/* Inputs - same as background, no contrast */
{sel} Input {{ background: {bg}; color: {text}; border: solid {border}; }}
{sel} Input:focus {{ border: solid {border}; background: {bg}; }}
{sel} Input.-invalid {{ border: solid {border}; background: {bg}; }}

/* TextArea - same as background, no contrast */
{sel} TextArea {{ background: {bg}; color: {text}; border: solid {border}; }}
{sel} TextArea:focus {{ border: solid {border}; background: {bg}; }}
{sel} TextArea .text-area--cursor-line {{ background: {hover}; }}
{sel} TextArea .text-area--cursor-gutter {{ background: {hover}; }}

/* Composer special - same as background */
{sel} #composer {{ background: {bg}; color: {text}; border: none; }}
{sel} #composer:focus {{ background: {bg}; border: none; }}
{sel} #composer .text-area--cursor-line {{ background: {hover}; }}
{sel} #composer_frame {{ background: {bg}; border-top: solid {border}; }}

/* Select widget - same as background */
{sel} Select {{ background: {bg}; color: {text}; border: solid {border}; }}
{sel} Select:focus {{ border: solid {border}; background: {bg}; }}
{sel} Select * {{ background: {bg}; color: {text}; }}
{sel} Select > * {{ background: {bg}; color: {text}; }}
{sel} Select > SelectCurrent {{ background: {bg}; color: {text}; }}
{sel} SelectCurrent {{ background: {bg}; color: {text}; }}
{sel} SelectCurrent * {{ background: {bg}; color: {text}; }}
{sel} SelectOverlay {{ background: {bg}; border: solid {border}; }}
{sel} SelectOverlay:focus {{ border: solid {border}; }}
{sel} OptionList {{ background: {bg}; }}
{sel} OptionList * {{ background: {bg}; }}
{sel} OptionList > .option-list--option {{ color: {text}; background: {bg}; }}
{sel} OptionList > .option-list--option-highlighted {{ background: {hover}; color: {text}; }}
{sel} OptionList > .option-list--option-hover {{ background: {hover}; }}

/* Settings tabs */
{sel} .settings_tab {{ background: {bg}; color: {muted}; }}
{sel} .settings_tab:hover {{ background: {hover}; color: {text}; }}
{sel} .settings_tab.active {{ background: {hover}; color: {text}; border-left: solid {border}; }}

/* Messages */
{sel} .msg {{ background: {bg}; color: {text}; }}
{sel} .msg:hover {{ background: {hover}; border-left: wide {accent}; }}
{sel} .msg.flash {{ background: {hover}; border-left: wide {accent}; }}
{sel} .msg_header {{ color: {text}; }}
{sel} .msg_body {{ color: {text}; }}
{sel} .msg_meta {{ color: {muted}; }}
{sel} .msg_reply {{ background: {bg}; color: {muted}; border-left: solid {accent}; }}

/* Composer */
{sel} #composer_frame {{ background: {bg}; border-top: solid {border}; }}
{sel} #reply_bar {{ color: {accent}; background: {bg}; }}
{sel} #send_btn {{ background: {accent_soft}; color: {text}; }}
{sel} #send_btn:hover {{ background: {accent_softer}; }}

/* Requests */
{sel} #req_in, {sel} #req_out {{ background: {bg}; border: round {border}; }}
{sel} #req_in_title, {sel} #req_out_title {{ color: {muted}; }}

/* Context menu */
{sel} #context_menu_container {{ background: {panel}; border: solid {accent}; }}
{sel} #menu_title {{ background: {accent}; color: {accent_text}; }}
{sel} #menu_items Button {{ background: {panel}; color: {text}; }}
{sel} #menu_items Button:hover {{ background: {accent}; color: {accent_text}; }}

/* Profile dialogs */
{sel} #profile_dialog_panel, {sel} #my_status_panel {{ background: {panel}; border: solid {border}; }}
{sel} .profile_bio_box {{ background: {bg}; border: solid {border}; }}

/* Scrollbars - CRITICAL for theme consistency */
{sel} * {{ scrollbar-background: {bg}; scrollbar-background-hover: {bg}; scrollbar-background-active: {bg}; scrollbar-color: {border}; scrollbar-color-hover: {border_light}; scrollbar-color-active: {border_light}; scrollbar-corner-color: {bg}; }}
{sel} #messages {{ scrollbar-background: {bg}; scrollbar-background-hover: {bg}; scrollbar-background-active: {bg}; scrollbar-color: {border}; scrollbar-color-hover: {border_light}; scrollbar-color-active: {border_light}; scrollbar-corner-color: {bg}; }}
{sel} .code-content-wrapper {{ scrollbar-background: {bg}; scrollbar-background-hover: {bg}; scrollbar-background-active: {bg}; scrollbar-color: {border}; scrollbar-color-hover: {border_light}; scrollbar-color-active: {border_light}; scrollbar-corner-color: {bg}; }}
{sel} VerticalScroll {{ scrollbar-background: {bg}; scrollbar-color: {border}; scrollbar-color-hover: {border_light}; }}
{sel} ScrollableContainer {{ scrollbar-background: {bg}; scrollbar-color: {border}; scrollbar-color-hover: {border_light}; }}
{sel} #settings_scroll {{ scrollbar-background: {bg}; scrollbar-color: {border}; }}

/* Checkboxes and switches */
{sel} Checkbox {{ color: {text}; }}
{sel} Checkbox > .toggle--button {{ background: {panel}; color: {muted}; }}
{sel} Checkbox.-on > .toggle--button {{ background: {accent}; color: {accent_text}; }}
{sel} Switch {{ background: {panel}; }}
{sel} Switch > .switch--slider {{ background: {muted}; }}
{sel} Switch.-on > .switch--slider {{ background: {accent}; }}

/* Progress bars */
{sel} ProgressBar > .bar--bar {{ background: {panel}; }}
{sel} ProgressBar > .bar--complete {{ background: {accent}; }}

/* DataTable */
{sel} DataTable {{ background: {bg}; }}
{sel} DataTable > .datatable--header {{ background: {panel}; color: {text}; }}
{sel} DataTable > .datatable--cursor {{ background: {hover}; }}
{sel} DataTable > .datatable--hover {{ background: {hover}; }}

/* Tree */
{sel} Tree {{ background: {bg}; }}
{sel} Tree > .tree--cursor {{ background: {hover}; }}
{sel} Tree > .tree--highlight {{ background: {hover}; }}
{sel} Tree > .tree--guides {{ color: {border}; }}

/* Labels and Static */
{sel} Label {{ color: {text}; }}
{sel} Static {{ color: {text}; }}
{sel} .section_label {{ color: {muted}; }}
{sel} .section_desc {{ color: {muted}; }}

/* Modal screens */
{sel} ModalScreen {{ background: {bg} 80%; }}

/* Tooltips */
{sel} Tooltip {{ background: {panel}; color: {text}; border: solid {border}; }}

/* Footer */
{sel} Footer {{ background: {panel}; color: {muted}; }}
{sel} Footer > .footer--key {{ background: {accent}; color: {accent_text}; }}

/* Header */
{sel} Header {{ background: {panel}; color: {text}; }}

/* Markdown */
{sel} Markdown {{ background: {bg}; color: {text}; }}
{sel} MarkdownH1, {sel} MarkdownH2, {sel} MarkdownH3 {{ color: {accent}; }}
{sel} MarkdownBlockQuote {{ background: {panel}; border-left: solid {accent}; }}

/* RichLog */
{sel} RichLog {{ background: {bg}; color: {text}; scrollbar-background: {bg}; scrollbar-color: {border}; }}

/* LoadingIndicator */
{sel} LoadingIndicator {{ background: {bg}; color: {accent}; }}

/* Placeholder */
{sel} Placeholder {{ background: {panel}; color: {muted}; border: dashed {border}; }}
"""



from textual.screen import Screen
from textual.widgets import Static, LoadingIndicator
from textual.containers import Vertical, Center
from textual.reactive import reactive

class LoadingScreen(Screen):
    status = reactive("Initializing...")

    CSS = """
    LoadingScreen {
        align: center middle;
        background: $surface;
    }
    #loading_container {
        width: 60;
        height: auto;
        border: solid $secondary;
        padding: 2;
        background: $panel;
    }
    #status_label {
        text-align: center;
        margin-top: 1;
        color: $text;
    }
    LoadingIndicator {
        height: 1;
        margin-bottom: 1;
    }
    """

    def compose(self):
        with Center():
            with Vertical(id="loading_container"):
                yield LoadingIndicator()
                yield Static(self.status, id="status_label")

    def watch_status(self, old_val, new_val):
        try:
            self.query_one("#status_label", Static).update(new_val)
        except:
            pass

    def update_status(self, text: str):
        self.status = text




from __future__ import annotations

import asyncio
import re
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import (
    Any, Callable, Dict, List, Optional, Type, Union,
    TYPE_CHECKING, Awaitable, Coroutine
)

if TYPE_CHECKING:
    from textual.app import App
    from textual.widget import Widget
    from textual.screen import Screen
    from textual.message import Message
    from textual.css.query import QueryType

class PluginSettingType(Enum):
    STRING = auto()
    INTEGER = auto()
    BOOLEAN = auto()
    SELECT = auto()
    COLOR = auto()
    KEYBIND = auto()
    TEXT = auto()
    PATH = auto()
    JSON = auto()
    LIST = auto()

@dataclass
class PluginSetting:
    key: str
    label: str
    type: PluginSettingType
    default: Any = None
    description: str = ""
    options: List[tuple] = field(default_factory=list)
    min_value: Optional[int] = None
    max_value: Optional[int] = None
    placeholder: str = ""
    required: bool = False
    hidden: bool = False

@dataclass
class PluginMeta:
    name: str
    id: str
    version: str = "1.0.0"
    description: str = ""
    author: str = "Unknown"
    website: str = ""
    requires_restart: bool = False
    tags: List[str] = field(default_factory=list)
    dependencies: List[str] = field(default_factory=list)
    min_app_version: str = ""
    license: str = "MIT"

class HookType(Enum):

    ON_APP_START = "on_app_start"
    ON_APP_STOP = "on_app_stop"
    ON_APP_READY = "on_app_ready"
    ON_CONNECT = "on_connect"
    ON_DISCONNECT = "on_disconnect"

    ON_MESSAGE_RECEIVE = "on_message_receive"
    ON_MESSAGE_SEND = "on_message_send"
    ON_MESSAGE_RENDER = "on_message_render"
    ON_MESSAGE_CONTEXT_MENU = "on_message_context_menu"
    ON_MESSAGE_DELETE = "on_message_delete"
    ON_MESSAGE_EDIT = "on_message_edit"

    ON_SIDEBAR_COMPOSE = "on_sidebar_compose"
    ON_COMPOSER_COMPOSE = "on_composer_compose"
    ON_CHAT_HEADER_COMPOSE = "on_chat_header_compose"
    ON_CHAT_FOOTER_COMPOSE = "on_chat_footer_compose"
    ON_TOOLBAR_COMPOSE = "on_toolbar_compose"

    ON_COMPOSE_SUBMIT = "on_compose_submit"
    ON_KEY_PRESS = "on_key_press"
    ON_KEY_RELEASE = "on_key_release"

    ON_CHAT_SWITCH = "on_chat_switch"
    ON_SCREEN_CHANGE = "on_screen_change"
    ON_TAB_CHANGE = "on_tab_change"

    ON_USER_CLICK = "on_user_click"
    ON_USER_JOIN = "on_user_join"
    ON_USER_LEAVE = "on_user_leave"
    ON_USER_STATUS_CHANGE = "on_user_status_change"

    ON_CALL_START = "on_call_start"
    ON_CALL_END = "on_call_end"
    ON_CALL_MUTE = "on_call_mute"

    ON_REQUEST_RECEIVE = "on_request_receive"
    ON_REQUEST_CLICK = "on_request_click"

    ON_COMMAND = "on_command"
    ON_SLASH_COMMAND = "on_slash_command"

    ON_SETTINGS_OPEN = "on_settings_open"
    ON_SETTINGS_SAVE = "on_settings_save"
    ON_SETTINGS_TAB_COMPOSE = "on_settings_tab_compose"

    ON_FILE_UPLOAD = "on_file_upload"
    ON_FILE_DOWNLOAD = "on_file_download"

    ON_THEME_CHANGE = "on_theme_change"

@dataclass
class RegisteredButton:
    id: str
    label: str
    callback: Callable
    icon: str = ""
    tooltip: str = ""
    location: str = "toolbar"
    variant: str = "default"
    classes: str = ""
    order: int = 100

@dataclass
class RegisteredPanel:
    id: str
    title: str
    widget_class: Type["Widget"]
    icon: str = ""
    location: str = "sidebar"
    width: Optional[int] = None
    height: Optional[int] = None
    collapsible: bool = True
    default_collapsed: bool = False

@dataclass
class RegisteredContextMenuItem:
    id: str
    label: str
    callback: Callable
    icon: str = ""
    target: str = "message"
    condition: Optional[Callable] = None
    order: int = 100

@dataclass
class RegisteredKeybinding:
    id: str
    key: str
    callback: Callable
    description: str = ""
    global_: bool = False
    context: str = ""

@dataclass
class RegisteredCommand:
    name: str
    callback: Callable
    description: str = ""
    usage: str = ""
    aliases: List[str] = field(default_factory=list)
    hidden: bool = False

@dataclass
class RegisteredWidget:
    id: str
    widget_class: Type["Widget"]
    location: str
    order: int = 100
    props: Dict[str, Any] = field(default_factory=dict)

@dataclass
class RegisteredSettingsTab:
    id: str
    title: str
    icon: str = ""
    widget_class: Optional[Type["Widget"]] = None
    compose_callback: Optional[Callable] = None
    order: int = 100

class Plugin(ABC):

    def __init__(self):
        self._app: Optional["App"] = None
        self._enabled: bool = False
        self._settings_values: Dict[str, Any] = {}
        self._settings_loaded: bool = False
        self._hooks: Dict[HookType, List[Callable]] = {}

        self._buttons: Dict[str, RegisteredButton] = {}
        self._panels: Dict[str, RegisteredPanel] = {}
        self._context_menu_items: Dict[str, RegisteredContextMenuItem] = {}
        self._keybindings: Dict[str, RegisteredKeybinding] = {}
        self._commands: Dict[str, RegisteredCommand] = {}
        self._widgets: Dict[str, RegisteredWidget] = {}
        self._settings_tabs: Dict[str, RegisteredSettingsTab] = {}
        self._mounted_widgets: List["Widget"] = []

        self._virtual_table_schemas: Dict[str, Dict[str, str]] = {}

    @property
    @abstractmethod
    def meta(self) -> PluginMeta:
        pass

    @property
    def settings(self) -> List[PluginSetting]:
        return []

    @property
    def css(self) -> str:
        return ""

    @property
    def app(self) -> Optional["App"]:
        return self._app

    @app.setter
    def app(self, value: "App") -> None:
        self._app = value

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._enabled = value

    @property
    def db(self) -> Any:
        if self._app and hasattr(self._app, 'db'):
            return self._app.db
        return None

    @property
    def network(self) -> Any:
        if self._app and hasattr(self._app, 'net'):
            return self._app.net
        return None

    @property
    def current_chat(self) -> Optional[str]:
        if self._app and hasattr(self._app, 'active_chat'):
            return self._app.active_chat
        return None

    @property
    def current_user(self) -> Optional[Dict[str, Any]]:
        if self._app and hasattr(self._app, 'me'):
            return self._app.me
        return None

    def get_lang(self) -> str:
        if self._app and hasattr(self._app, 'lang'):
            return self._app.lang or "en"
        return "en"

    @property
    def translations(self) -> dict[str, dict[str, str]]:
        return {}

    def t(self, key: str, **kwargs) -> str:
        lang = self.get_lang()

        plugin_trans = getattr(self, '_plugin_translations', None)
        if not plugin_trans:
            try:
                plugin_trans = self.translations or {}
            except Exception:
                plugin_trans = {}

        if plugin_trans:

            if lang in plugin_trans:
                lang_trans = plugin_trans[lang]
                if key in lang_trans:
                    try:
                        return lang_trans[key].format(**kwargs) if kwargs else lang_trans[key]
                    except (KeyError, ValueError):
                        return lang_trans[key]

            if "en" in plugin_trans:
                en_trans = plugin_trans["en"]
                if key in en_trans:
                    try:
                        return en_trans[key].format(**kwargs) if kwargs else en_trans[key]
                    except (KeyError, ValueError):
                        return en_trans[key]

        try:
            from ui.i18n import t as global_translate
            return global_translate(lang, key, **kwargs)
        except Exception:
            return key

    def translate(self, key: str, **kwargs) -> str:
        return self.t(key, **kwargs)

    def _ensure_settings_loaded(self) -> None:
        if self._settings_loaded:
            return
        try:
            from plugins.loader import get_plugin_loader
            loader = get_plugin_loader()
            if loader:
                saved = loader.get_plugin_settings(self.meta.id)
                if saved:
                    self._settings_values = saved.copy()
        except Exception:
            pass
        self._settings_loaded = True

    def get_setting(self, key: str, default: Any = None) -> Any:

        self._ensure_settings_loaded()

        if key in self._settings_values:
            return self._settings_values[key]

        for setting in self.settings:
            if setting.key == key:
                return setting.default if default is None else default
        return default

    def set_setting(self, key: str, value: Any) -> None:
        self._ensure_settings_loaded()
        self._settings_values[key] = value
        self._save_settings()

    def load_settings(self, values: Dict[str, Any]) -> None:
        self._settings_values = values.copy()
        self._settings_loaded = True

    def reload_settings(self) -> None:
        self._settings_loaded = False
        self._settings_values = {}
        self._ensure_settings_loaded()

    def save_settings(self) -> Dict[str, Any]:
        return self._settings_values.copy()

    def _save_settings(self) -> None:
        from plugins.loader import get_plugin_loader
        loader = get_plugin_loader()
        loader.save_plugin_settings(self.meta.id, self._settings_values)

    def on_load(self) -> None:
        pass

    def on_unload(self) -> None:
        self._cleanup()

    def on_enable(self) -> None:
        pass

    def on_disable(self) -> None:
        self._cleanup()

    def _cleanup(self) -> None:
        for widget in self._mounted_widgets:
            try:
                widget.remove()
            except Exception:
                pass
        self._mounted_widgets.clear()

        self._close_plugin_db()

    def register_hook(self, hook_type: HookType, callback: Callable) -> None:
        if hook_type not in self._hooks:
            self._hooks[hook_type] = []
        self._hooks[hook_type].append(callback)

    def unregister_hook(self, hook_type: HookType, callback: Callable) -> None:
        if hook_type in self._hooks:
            try:
                self._hooks[hook_type].remove(callback)
            except ValueError:
                pass

    def get_hooks(self, hook_type: HookType) -> List[Callable]:
        return self._hooks.get(hook_type, [])

    def register_button(
        self,
        id: str,
        label: str,
        callback: Callable,
        icon: str = "",
        tooltip: str = "",
        location: str = "toolbar",
        variant: str = "default",
        classes: str = "",
        order: int = 100
    ) -> RegisteredButton:
        btn = RegisteredButton(
            id=f"{self.meta.id}_{id}",
            label=label,
            callback=callback,
            icon=icon,
            tooltip=tooltip,
            location=location,
            variant=variant,
            classes=classes,
            order=order
        )
        self._buttons[id] = btn
        return btn

    def unregister_button(self, id: str) -> None:
        self._buttons.pop(id, None)

    def get_buttons(self, location: Optional[str] = None) -> List[RegisteredButton]:
        buttons = list(self._buttons.values())
        if location:
            buttons = [b for b in buttons if b.location == location]
        return sorted(buttons, key=lambda b: b.order)

    def register_panel(
        self,
        id: str,
        title: str,
        widget_class: Type["Widget"],
        icon: str = "",
        location: str = "sidebar",
        width: Optional[int] = None,
        height: Optional[int] = None,
        collapsible: bool = True,
        default_collapsed: bool = False
    ) -> RegisteredPanel:
        panel = RegisteredPanel(
            id=f"{self.meta.id}_{id}",
            title=title,
            widget_class=widget_class,
            icon=icon,
            location=location,
            width=width,
            height=height,
            collapsible=collapsible,
            default_collapsed=default_collapsed
        )
        self._panels[id] = panel
        return panel

    def unregister_panel(self, id: str) -> None:
        self._panels.pop(id, None)

    def get_panels(self, location: Optional[str] = None) -> List[RegisteredPanel]:
        panels = list(self._panels.values())
        if location:
            panels = [p for p in panels if p.location == location]
        return panels

    def register_context_menu_item(
        self,
        id: str,
        label: str,
        callback: Callable,
        icon: str = "",
        target: str = "message",
        condition: Optional[Callable] = None,
        order: int = 100
    ) -> RegisteredContextMenuItem:
        item = RegisteredContextMenuItem(
            id=f"{self.meta.id}_{id}",
            label=label,
            callback=callback,
            icon=icon,
            target=target,
            condition=condition,
            order=order
        )
        self._context_menu_items[id] = item
        return item

    def get_context_menu_items(self, target: Optional[str] = None) -> List[RegisteredContextMenuItem]:
        items = list(self._context_menu_items.values())
        if target:
            items = [i for i in items if i.target == target]
        return sorted(items, key=lambda i: i.order)

    def register_keybinding(
        self,
        id: str,
        key: str,
        callback: Callable,
        description: str = "",
        global_: bool = False,
        context: str = ""
    ) -> RegisteredKeybinding:
        kb = RegisteredKeybinding(
            id=f"{self.meta.id}_{id}",
            key=key,
            callback=callback,
            description=description,
            global_=global_,
            context=context
        )
        self._keybindings[id] = kb
        return kb

    def get_keybindings(self) -> List[RegisteredKeybinding]:
        return list(self._keybindings.values())

    def register_command(
        self,
        name: str,
        callback: Callable,
        description: str = "",
        usage: str = "",
        aliases: Optional[List[str]] = None,
        hidden: bool = False
    ) -> RegisteredCommand:
        cmd = RegisteredCommand(
            name=name,
            callback=callback,
            description=description,
            usage=usage or f"/{name}",
            aliases=aliases or [],
            hidden=hidden
        )
        self._commands[name] = cmd
        return cmd

    def get_commands(self) -> List[RegisteredCommand]:
        return [c for c in self._commands.values() if not c.hidden]

    def get_all_commands(self) -> Dict[str, RegisteredCommand]:
        return self._commands.copy()

    def register_widget(
        self,
        id: str,
        widget_class: Type["Widget"],
        location: str,
        order: int = 100,
        **props
    ) -> RegisteredWidget:
        widget = RegisteredWidget(
            id=f"{self.meta.id}_{id}",
            widget_class=widget_class,
            location=location,
            order=order,
            props=props
        )
        self._widgets[id] = widget
        return widget

    def get_widgets(self, location: Optional[str] = None) -> List[RegisteredWidget]:
        widgets = list(self._widgets.values())
        if location:
            widgets = [w for w in widgets if w.location == location]
        return sorted(widgets, key=lambda w: w.order)

    def register_settings_tab(
        self,
        id: str,
        title: str,
        icon: str = "",
        widget_class: Optional[Type["Widget"]] = None,
        compose_callback: Optional[Callable] = None,
        order: int = 100
    ) -> RegisteredSettingsTab:
        tab = RegisteredSettingsTab(
            id=f"{self.meta.id}_{id}",
            title=title,
            icon=icon,
            widget_class=widget_class,
            compose_callback=compose_callback,
            order=order
        )
        self._settings_tabs[id] = tab
        return tab

    def get_settings_tabs(self) -> List[RegisteredSettingsTab]:
        return sorted(self._settings_tabs.values(), key=lambda t: t.order)

    def log(self, message: str) -> None:
        print(f"[{self.meta.name}] {message}")

    def notify(self, message: str, severity: str = "information", timeout: Optional[float] = None) -> None:
        if self._app:
            if timeout is None:

                base_timeout = 5.0
                extra_time = len(message) / 50.0
                timeout = min(base_timeout + extra_time, 30.0)

                if severity == "error":
                    timeout = max(timeout, 10.0)
                elif severity == "warning":
                    timeout = max(timeout, 7.0)

            self._app.notify(message, severity=severity, timeout=timeout)

    def notify_persistent(self, message: str, severity: str = "information") -> None:
        if self._app:
            self._app.notify(message, severity=severity, timeout=0)

    def show_toast(self, message: str, timeout: float = 2.0) -> None:
        self.notify(message, timeout=timeout)

    def show_dialog(
        self,
        title: str,
        content: Union[str, "Widget"],
        buttons: Optional[List[tuple]] = None,
        callback: Optional[Callable] = None
    ) -> None:
        if not self._app:
            return
        from ui.dialogs import PluginDialog
        dialog = PluginDialog(
            title=title,
            content=content,
            buttons=buttons or [("OK", "ok")],
            plugin_id=self.meta.id
        )
        self._app.push_screen(dialog, callback)

    def show_confirm(
        self,
        title: str,
        message: str,
        callback: Callable[[bool], None]
    ) -> None:
        def on_result(result):
            callback(result == "yes")
        self.show_dialog(
            title=title,
            content=message,
            buttons=[("Yes", "yes"), ("No", "no")],
            callback=on_result
        )

    def show_input(
        self,
        title: str,
        prompt: str,
        default: str = "",
        callback: Optional[Callable[[Optional[str]], None]] = None
    ) -> None:
        if not self._app:
            return
        from ui.dialogs import PluginInputDialog
        dialog = PluginInputDialog(
            title=title,
            prompt=prompt,
            default=default,
            plugin_id=self.meta.id
        )
        self._app.push_screen(dialog, callback)

    def mount_widget(
        self,
        widget: "Widget",
        target: Optional[Union[str, "Widget"]] = None,
        before: Optional["Widget"] = None,
        after: Optional["Widget"] = None
    ) -> None:
        if not self._app:
            return
        try:
            if target is None:
                self._app.mount(widget, before=before, after=after)
            elif isinstance(target, str):
                container = self._app.query_one(target)
                container.mount(widget, before=before, after=after)
            else:
                target.mount(widget, before=before, after=after)
            self._mounted_widgets.append(widget)
        except Exception as e:
            self.log(f"Error mounting widget: {e}")

    async def mount_widget_async(
        self,
        widget: "Widget",
        target: Optional[Union[str, "Widget"]] = None
    ) -> None:
        if not self._app:
            return
        try:
            if target is None:
                await self._app.mount(widget)
            elif isinstance(target, str):
                container = self._app.query_one(target)
                await container.mount(widget)
            else:
                await target.mount(widget)
            self._mounted_widgets.append(widget)
        except Exception as e:
            self.log(f"Error mounting widget: {e}")

    def unmount_widget(self, widget: "Widget") -> None:
        try:
            widget.remove()
            if widget in self._mounted_widgets:
                self._mounted_widgets.remove(widget)
        except Exception as e:
            self.log(f"Error unmounting widget: {e}")

    def query(self, selector: str):
        if not self._app:
            return []
        return self._app.query(selector)

    def query_one(self, selector: str, expect_type: Optional[Type] = None) -> Optional["Widget"]:
        if not self._app:
            return None
        try:
            if expect_type:
                return self._app.query_one(selector, expect_type)
            return self._app.query_one(selector)
        except Exception:
            return None

    def push_screen(self, screen: "Screen", callback: Optional[Callable] = None) -> None:
        if self._app:
            self._app.push_screen(screen, callback)

    def pop_screen(self) -> None:
        if self._app:
            self._app.pop_screen()

    def switch_screen(self, screen: Union[str, "Screen"]) -> None:
        if self._app:
            self._app.switch_screen(screen)

    def send_message(self, content: str, chat_id: Optional[str] = None) -> None:
        if not self._app:
            return
        target_chat = chat_id or self.current_chat
        if not target_chat:
            self.log("No chat selected")
            return
        if hasattr(self._app, 'send_message'):
            self._app.send_message(content, target_chat)
        elif hasattr(self._app, '_do_send'):
            asyncio.create_task(self._app._do_send(content))

    def get_messages(self, chat_id: Optional[str] = None, limit: int = 50) -> List[Dict]:
        if not self.db:
            return []
        target_chat = chat_id or self.current_chat
        if not target_chat:
            return []
        try:
            return list(self.db.list_messages(target_chat, limit=limit))
        except Exception:
            return []

    def jump_to_message(self, message_id: str) -> bool:
        if not self._app or not message_id:
            return False

        msg_widgets = getattr(self._app, '_msg_widgets', {})
        if message_id not in msg_widgets:
            return False

        try:
            from ui.widgets.messaging import MessageWidget
            w = msg_widgets[message_id]
            if isinstance(w, MessageWidget):
                w.post_message(MessageWidget.JumpRequested(w, message_id))
                return True
        except Exception:
            pass

        return False

    def _get_db(self):
        if self._app and hasattr(self._app, 'db') and self._app.db:
            return self._app.db
        return None

    def _generate_row_id(self, data: Dict[str, Any]) -> str:

        for pk_field in ('id', 'msg_id', 'message_id', 'key', 'row_id'):
            if pk_field in data and data[pk_field] is not None:
                return str(data[pk_field])
        return str(uuid.uuid4())

    def plugin_db_create_table(self, table_name: str, columns: Dict[str, str]) -> None:

        if not hasattr(self, '_virtual_table_schemas'):
            self._virtual_table_schemas = {}
        self._virtual_table_schemas[table_name] = columns

    def plugin_db_insert(self, table_name: str, data: Dict[str, Any]) -> None:
        db = self._get_db()
        if not db:
            self.log("Warning: No database available for plugin_db_insert")
            return

        row_id = self._generate_row_id(data)
        db.plugin_data_insert(self.meta.id, table_name, row_id, data)

    def plugin_db_select(
        self, table_name: str, where: str = "", params: tuple = (), limit: int = 0
    ) -> List[Dict[str, Any]]:
        db = self._get_db()
        if not db:
            return []

        if not where:

            return db.plugin_data_select(self.meta.id, table_name, limit=limit)

        import re

        simple_match = re.match(r'^(\w+)\s*=\s*\?$', where.strip())
        if simple_match and len(params) == 1:
            field = simple_match.group(1)
            return db.plugin_data_select(
                self.meta.id, table_name,
                where_field=field, where_value=params[0],
                limit=limit
            )

        try:

            json_where_parts = []
            pattern = re.compile(r'(\w+)\s*([=<>!]+)\s*\?')

            for match in pattern.finditer(where):
                field = match.group(1)
                op = match.group(2)
                json_where_parts.append(f"JSON_EXTRACT(data_json, '$.{field}') {op} ?")

            if json_where_parts:

                json_where = where
                for match in pattern.finditer(where):
                    field = match.group(1)
                    op = match.group(2)
                    old_expr = f"{field} {op} ?"
                    new_expr = f"JSON_EXTRACT(data_json, '$.{field}') {op} ?"
                    json_where = json_where.replace(old_expr, new_expr, 1)

                result = db.plugin_data_execute_json_query(
                    self.meta.id, table_name, json_where, params
                )
                if limit > 0:
                    return result[:limit]
                return result
        except Exception as e:
            self.log(f"Warning: Complex WHERE parsing failed: {e}")

        all_rows = db.plugin_data_select(self.meta.id, table_name)

        if where and params:
            import re
            matches = re.findall(r'(\w+)\s*=\s*\?', where)
            if matches and len(matches) == len(params):
                filtered = []
                for row in all_rows:
                    match = True
                    for field, value in zip(matches, params):
                        if str(row.get(field, "")) != str(value):
                            match = False
                            break
                    if match:
                        filtered.append(row)
                return filtered[:limit] if limit > 0 else filtered

        return all_rows[:limit] if limit > 0 else all_rows

    def plugin_db_delete(self, table_name: str, where: str, params: tuple = ()) -> int:
        db = self._get_db()
        if not db:
            return 0

        import re

        simple_match = re.match(r'^(\w+)\s*=\s*\?$', where.strip())
        if simple_match and len(params) == 1:
            field = simple_match.group(1)

            if field in ('id', 'msg_id', 'message_id', 'key', 'row_id'):
                return db.plugin_data_delete(self.meta.id, table_name, str(params[0]))

            return db.plugin_data_delete_where(self.meta.id, table_name, field, params[0])

        rows = self.plugin_db_select(table_name, where, params)
        count = 0
        for row in rows:
            row_id = self._generate_row_id(row)
            count += db.plugin_data_delete(self.meta.id, table_name, row_id)
        return count

    def plugin_db_update(
        self, table_name: str, data: Dict[str, Any], where: str, params: tuple = ()
    ) -> int:
        db = self._get_db()
        if not db:
            return 0

        rows = self.plugin_db_select(table_name, where, params)
        count = 0

        for row in rows:

            updated_row = {**row, **data}
            row_id = self._generate_row_id(updated_row)
            db.plugin_data_insert(self.meta.id, table_name, row_id, updated_row)
            count += 1

        return count

    def plugin_db_count(self, table_name: str, where: str = "", params: tuple = ()) -> int:
        db = self._get_db()
        if not db:
            return 0

        import re

        if not where:
            return db.plugin_data_count(self.meta.id, table_name)

        simple_match = re.match(r'^(\w+)\s*=\s*\?$', where.strip())
        if simple_match and len(params) == 1:
            field = simple_match.group(1)
            return db.plugin_data_count(self.meta.id, table_name, field, params[0])

        return len(self.plugin_db_select(table_name, where, params))

    def plugin_db_clear_table(self, table_name: str) -> int:
        db = self._get_db()
        if not db:
            return 0
        return db.plugin_data_clear_table(self.meta.id, table_name)

    def plugin_db_clear_all(self) -> int:
        db = self._get_db()
        if not db:
            return 0
        return db.plugin_data_clear_all(self.meta.id)

    def db_get(self, key: str, default: Any = None) -> Any:
        db = self._get_db()
        if not db:
            return default

        try:
            data = db.plugin_data_get(self.meta.id, "kv_store", key)
            if data and "value" in data:
                return data["value"]
            return default
        except Exception:
            return default

    def db_set(self, key: str, value: Any) -> None:
        db = self._get_db()
        if not db:
            self.log("Warning: No database available for db_set")
            return

        try:
            db.plugin_data_insert(
                self.meta.id, "kv_store", key,
                {"key": key, "value": str(value) if value is not None else None}
            )
        except Exception as e:
            self.log(f"Error saving to db: {e}")

    def db_delete(self, key: str) -> None:
        db = self._get_db()
        if not db:
            return

        try:
            db.plugin_data_delete(self.meta.id, "kv_store", key)
        except Exception:
            pass

    def db_list_keys(self, prefix: str = "") -> List[str]:
        db = self._get_db()
        if not db:
            return []

        try:
            rows = db.plugin_data_select(self.meta.id, "kv_store")
            keys = [row.get("key", "") for row in rows if row.get("key")]
            if prefix:
                keys = [k for k in keys if k.startswith(prefix)]
            return keys
        except Exception:
            return []

    def _close_plugin_db(self) -> None:

        pass

    def plugin_db_execute(self, sql: str, params: tuple = ()) -> None:
        self.log(f"Warning: plugin_db_execute() is deprecated and has no effect. SQL: {sql[:50]}...")

    def plugin_db_commit(self) -> None:
        pass

    def call_later(self, callback: Callable, delay: float = 0) -> None:
        if self._app:
            if delay > 0:
                self._app.call_later(lambda: asyncio.create_task(self._delayed_call(callback, delay)))
            else:
                self._app.call_later(callback)

    async def _delayed_call(self, callback: Callable, delay: float) -> None:
        await asyncio.sleep(delay)
        callback()

    def run_async(self, coro: Coroutine) -> None:
        asyncio.create_task(coro)

    def t(self, key: str, **kwargs) -> str:
        if self._app and hasattr(self._app, 't'):
            return self._app.t(key, **kwargs)
        return key

    def inject_css(self, css: str) -> None:
        if self._app:
            if not hasattr(self._app, '_plugin_css'):
                self._app._plugin_css = {}
            self._app._plugin_css[self.meta.id] = css
            try:
                self._app.refresh_css()
            except Exception:
                pass

    def remove_css(self) -> None:
        if self._app and hasattr(self._app, '_plugin_css'):
            self._app._plugin_css.pop(self.meta.id, None)
            try:
                self._app.refresh_css()
            except Exception:
                pass

    def on_message_receive(self, sender: str, message: str, meta: dict) -> Optional[str]:
        return None

    def on_message_send(self, message: str) -> Optional[str]:
        return None

    def on_message_render(self, sender: str, message: str, meta: dict) -> Optional[str]:
        return None

    def on_command(self, command: str, args: List[str]) -> bool:
        if command in self._commands:
            try:
                self._commands[command].callback(args)
                return True
            except Exception as e:
                self.log(f"Error in command {command}: {e}")
        return False

    def on_key_press(self, key: str, modifiers: List[str]) -> bool:
        key_combo = "+".join(modifiers + [key]) if modifiers else key
        for kb in self._keybindings.values():
            if kb.key.lower() == key_combo.lower():
                try:
                    kb.callback()
                    return True
                except Exception as e:
                    self.log(f"Error in keybinding {kb.id}: {e}")
        return False

    def on_app_ready(self) -> None:
        if self.css:
            self.inject_css(self.css)

class PluginError(Exception):
    pass

class PluginLoadError(PluginError):
    pass

class PluginSettingsError(PluginError):
    pass

class PluginDependencyError(PluginError):
    pass
